from __future__ import annotations

from datetime import timedelta
from typing import List, Union

from pydantic import Field, RootModel
from typing_extensions import Literal

from datahub.api.entities.datacontract.assertion import BaseAssertion
from datahub.emitter.mce_builder import make_assertion_source
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.metadata.schema_classes import (
    AssertionInfoClass,
    AssertionTypeClass,
    CalendarIntervalClass,
    FixedIntervalScheduleClass,
    FreshnessAssertionInfoClass,
    FreshnessAssertionScheduleClass,
    FreshnessAssertionScheduleTypeClass,
    FreshnessAssertionTypeClass,
    FreshnessCronScheduleClass,
)


class CronFreshnessAssertion(BaseAssertion):
    type: Literal["cron"]

    cron: str = Field(
        description="The cron expression to use. See https://crontab.guru/ for help."
    )
    timezone: str = Field(
        "UTC",
        description="The timezone to use for the cron schedule. Defaults to UTC.",
    )

    def generate_freshness_assertion_schedule(self) -> FreshnessAssertionScheduleClass:
        return FreshnessAssertionScheduleClass(
            type=FreshnessAssertionScheduleTypeClass.CRON,
            cron=FreshnessCronScheduleClass(
                cron=self.cron,
                timezone=self.timezone,
            ),
        )


class FixedIntervalFreshnessAssertion(BaseAssertion):
    type: Literal["interval"]

    interval: timedelta

    def generate_freshness_assertion_schedule(self) -> FreshnessAssertionScheduleClass:
        return FreshnessAssertionScheduleClass(
            type=FreshnessAssertionScheduleTypeClass.FIXED_INTERVAL,
            fixedInterval=FixedIntervalScheduleClass(
                unit=CalendarIntervalClass.SECOND,
                multiple=int(self.interval.total_seconds()),
            ),
        )


class FreshnessAssertion(
    RootModel[Union[CronFreshnessAssertion, FixedIntervalFreshnessAssertion]]
):
    root: Union[CronFreshnessAssertion, FixedIntervalFreshnessAssertion] = Field(
        discriminator="type"
    )

    @property
    def id(self):
        return self.root.type

    def generate_mcp(
        self, assertion_urn: str, entity_urn: str
    ) -> List[MetadataChangeProposalWrapper]:
        aspect = AssertionInfoClass(
            type=AssertionTypeClass.FRESHNESS,
            freshnessAssertion=FreshnessAssertionInfoClass(
                entity=entity_urn,
                type=FreshnessAssertionTypeClass.DATASET_CHANGE,
                schedule=self.root.generate_freshness_assertion_schedule(),
            ),
            source=make_assertion_source(),
            description=self.root.description,
        )
        return [MetadataChangeProposalWrapper(entityUrn=assertion_urn, aspect=aspect)]



import contextlib
import dataclasses
import functools
import json
import logging
import pathlib
import re
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import cached_property
from typing import Any, Dict, Iterable, List, Optional, Union

import pydantic
from typing_extensions import Self

from datahub.configuration.common import AllowDenyPattern, ConfigModel, HiddenFromDocs
from datahub.configuration.time_window_config import (
    BaseTimeWindowConfig,
    BucketDuration,
    get_time_bucket,
)
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import SupportStatus, config_class, support_status
from datahub.ingestion.api.report import Report
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.source_helpers import auto_workunit
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.source.snowflake.constants import SnowflakeObjectDomain
from datahub.ingestion.source.snowflake.snowflake_config import (
    DEFAULT_TEMP_TABLES_PATTERNS,
    QueryDedupStrategyType,
    SnowflakeFilterConfig,
    SnowflakeIdentifierConfig,
)
from datahub.ingestion.source.snowflake.snowflake_connection import (
    SnowflakeConnection,
    SnowflakeConnectionConfig,
)
from datahub.ingestion.source.snowflake.snowflake_lineage_v2 import (
    SnowflakeLineageExtractor,
)
from datahub.ingestion.source.snowflake.snowflake_query import SnowflakeQuery
from datahub.ingestion.source.snowflake.snowflake_utils import (
    SnowflakeFilter,
    SnowflakeIdentifierBuilder,
    SnowflakeStructuredReportMixin,
)
from datahub.ingestion.source.snowflake.stored_proc_lineage import (
    StoredProcCall,
    StoredProcLineageReport,
    StoredProcLineageTracker,
)
from datahub.ingestion.source.state.redundant_run_skip_handler import (
    RedundantQueriesRunSkipHandler,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig
from datahub.metadata.urns import CorpUserUrn
from datahub.sql_parsing.schema_resolver import SchemaResolver
from datahub.sql_parsing.sql_parsing_aggregator import (
    KnownLineageMapping,
    ObservedQuery,
    PreparsedQuery,
    SqlAggregatorReport,
    SqlParsingAggregator,
    TableRename,
    TableSwap,
    UrnStr,
)
from datahub.sql_parsing.sql_parsing_common import QueryType
from datahub.sql_parsing.sqlglot_lineage import (
    ColumnLineageInfo,
    ColumnRef,
    DownstreamColumnRef,
)
from datahub.sql_parsing.sqlglot_utils import get_query_fingerprint
from datahub.utilities.file_backed_collections import (
    ConnectionWrapper,
    FileBackedList,
)
from datahub.utilities.lossy_collections import LossyList
from datahub.utilities.perf_timer import PerfTimer

logger = logging.getLogger(__name__)

_SYS_VIEW_PLACEHOLDER_PREFIX = "$SYS_VIEW_"


@dataclass(frozen=True)
class _AuditLogObjectClassification:
    """Summary of name anomalies found in one audit-log object list."""

    has_placeholder: bool  # any $SYS_VIEW_* name; blocks fast path
    n_missing_name: int  # None/""/whitespace-only objectNames; metrics only
    n_unknown_dollar: int  # non-$SYS_VIEW_ dollar prefix; blocks fast path

    @property
    def has_unknown_dollar_prefix(self) -> bool:
        return self.n_unknown_dollar > 0


@dataclass(frozen=True)
class _DownstreamTarget:
    """One downstream entry from a Snowflake audit log row.

    downstream and column_lineage are both None for usage-only queries
    (e.g. SELECTs), and both non-None for queries with a downstream table.
    """

    downstream: Optional[UrnStr]
    column_lineage: Optional[List[ColumnLineageInfo]]
    secondary_id: Optional[str]

    def __post_init__(self) -> None:
        if (self.downstream is None) != (self.column_lineage is None):
            raise ValueError(
                "downstream and column_lineage must be both None or both non-None"
            )


# Define a type alias
UserName = str
UserEmail = str
UsersMapping = Dict[UserName, UserEmail]


class SnowflakeQueriesExtractorConfig(ConfigModel):
    # TODO: Support stateful ingestion for the time windows.
    window: BaseTimeWindowConfig = BaseTimeWindowConfig()

    pushdown_deny_usernames: List[str] = pydantic.Field(
        default=[],
        description="List of snowflake usernames (SQL LIKE patterns, e.g., 'SERVICE_%', '%_PROD', 'TEST_USER') which will NOT be considered for lineage/usage/queries extraction. "
        "This is primarily useful for improving performance by filtering out users with extremely high query volumes.",
    )

    pushdown_allow_usernames: List[str] = pydantic.Field(
        default=[],
        description="List of snowflake usernames (SQL LIKE patterns, e.g., 'ANALYST_%', '%_USER', 'MAIN_ACCOUNT') which WILL be considered for lineage/usage/queries extraction. "
        "This is primarily useful for improving performance by filtering in only specific users. "
        "If not specified, all users not in deny list are included.",
    )

    user_email_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for user emails to filter in usage.",
    )

    temporary_tables_pattern: List[str] = pydantic.Field(
        default=DEFAULT_TEMP_TABLES_PATTERNS,
        description="[Advanced] Regex patterns for temporary tables to filter in lineage ingestion. Specify regex to "
        "match the entire table name in database.schema.table format. Defaults are to set in such a way "
        "to ignore the temporary staging tables created by known ETL tools.",
    )

    local_temp_path: HiddenFromDocs[Optional[pathlib.Path]] = pydantic.Field(
        # TODO: For now, this is simply an advanced config to make local testing easier.
        # Eventually, we will want to store date-specific files in the directory and use it as a cache.
        default=None,
        description="Local path to store the audit log.",
    )

    include_lineage: bool = True
    include_queries: bool = True
    include_usage_statistics: bool = True
    include_query_usage_statistics: bool = True
    include_operations: bool = True

    push_down_database_pattern_access_history: bool = pydantic.Field(
        default=False,
        description="If enabled, pushes down database pattern filtering to the access_history table for improved performance. "
        "This filters on the accessed objects in access_history.",
    )

    additional_database_names_allowlist: List[str] = pydantic.Field(
        default=[],
        description="Additional database names (no pattern matching) to be included in the access_history filter. "
        "Only applies if push_down_database_pattern_access_history=True. "
        "These databases will be included in the filter being pushed down regardless of database_pattern settings."
        "This may be required in the case of _eg_ temporary tables being created in a different database than the ones in the database_name patterns.",
    )

    query_dedup_strategy: QueryDedupStrategyType = QueryDedupStrategyType.STANDARD

    @cached_property
    def _compiled_temporary_tables_pattern(self) -> "List[re.Pattern[str]]":
        return [
            re.compile(pattern, re.IGNORECASE)
            for pattern in self.temporary_tables_pattern
        ]


class SnowflakeQueriesSourceConfig(
    SnowflakeQueriesExtractorConfig, SnowflakeIdentifierConfig, SnowflakeFilterConfig
):
    connection: SnowflakeConnectionConfig


@dataclass
class SnowflakeQueriesExtractorReport(Report):
    copy_history_fetch_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    query_log_fetch_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    users_fetch_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)

    audit_log_load_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    aggregator_generate_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)

    sql_aggregator: Optional[SqlAggregatorReport] = None
    stored_proc_lineage: Optional[StoredProcLineageReport] = None

    num_ddl_queries_dropped: int = 0
    num_stream_queries_observed: int = 0
    num_stream_queries_clean_fast_path: int = 0
    num_create_temp_view_queries_observed: int = 0
    num_audit_rows_missing_object_name: int = 0
    num_audit_rows_unknown_dollar_prefix: int = 0
    num_users: int = 0
    num_queries_with_empty_column_name: int = 0
    queries_with_empty_column_name: LossyList[str] = dataclasses.field(
        default_factory=LossyList
    )


@dataclass
class SnowflakeQueriesSourceReport(SourceReport):
    window: Optional[BaseTimeWindowConfig] = None
    queries_extractor: Optional[SnowflakeQueriesExtractorReport] = None


class SnowflakeQueriesExtractor(SnowflakeStructuredReportMixin, Closeable):
    def __init__(
        self,
        connection: SnowflakeConnection,
        config: SnowflakeQueriesExtractorConfig,
        structured_report: SourceReport,
        filters: SnowflakeFilter,
        identifiers: SnowflakeIdentifierBuilder,
        redundant_run_skip_handler: Optional[RedundantQueriesRunSkipHandler] = None,
        graph: Optional[DataHubGraph] = None,
        schema_resolver: Optional[SchemaResolver] = None,
        discovered_tables: Optional[List[str]] = None,
    ):
        self.connection = connection

        self.config = config
        self.report = SnowflakeQueriesExtractorReport()
        self.filters = filters
        self.identifiers = identifiers
        self.discovered_tables = set(discovered_tables) if discovered_tables else None
        self.redundant_run_skip_handler = redundant_run_skip_handler

        self._structured_report = structured_report

        # Adjust time window based on stateful ingestion state
        self.start_time, self.end_time = self._get_time_window()

        # The exit stack helps ensure that we close all the resources we open.
        self._exit_stack = contextlib.ExitStack()

        self.aggregator: SqlParsingAggregator = self._exit_stack.enter_context(
            SqlParsingAggregator(
                platform=self.identifiers.platform,
                platform_instance=self.identifiers.identifier_config.platform_instance,
                env=self.identifiers.identifier_config.env,
                schema_resolver=schema_resolver,
                graph=graph,
                eager_graph_load=False,
                generate_lineage=self.config.include_lineage,
                generate_queries=self.config.include_queries,
                generate_usage_statistics=self.config.include_usage_statistics,
                generate_query_usage_statistics=self.config.include_query_usage_statistics,
                usage_config=BaseUsageConfig(
                    bucket_duration=self.config.window.bucket_duration,
                    start_time=self.start_time,
                    end_time=self.end_time,
                    user_email_pattern=self.config.user_email_pattern,
                    # TODO make the rest of the fields configurable
                ),
                generate_operations=self.config.include_operations,
                is_temp_table=self.is_temp_table,
                is_allowed_table=self.is_allowed_table,
                format_queries=False,
            )
        )
        self.report.sql_aggregator = self.aggregator.report

    @property
    def structured_reporter(self) -> SourceReport:
        return self._structured_report

    def _get_time_window(self) -> tuple[datetime, datetime]:
        if self.redundant_run_skip_handler:
            start_time, end_time = (
                self.redundant_run_skip_handler.suggest_run_time_window(
                    self.config.window.start_time,
                    self.config.window.end_time,
                )
            )
        else:
            start_time = self.config.window.start_time
            end_time = self.config.window.end_time

        # Usage statistics are aggregated per bucket (typically per day).
        # To ensure accurate aggregated metrics, we need to align the start_time
        # to the beginning of a bucket so that we include complete bucket periods.
        if self.config.include_usage_statistics:
            start_time = get_time_bucket(start_time, self.config.window.bucket_duration)

        return start_time, end_time

    def _update_state(self) -> None:
        if self.redundant_run_skip_handler:
            self.redundant_run_skip_handler.update_state(
                self.config.window.start_time,
                self.config.window.end_time,
                self.config.window.bucket_duration,
            )

    @functools.cached_property
    def local_temp_path(self) -> pathlib.Path:
        if self.config.local_temp_path:
            assert self.config.local_temp_path.is_dir()
            return self.config.local_temp_path

        path = pathlib.Path(tempfile.mkdtemp())
        path.mkdir(parents=True, exist_ok=True)
        logger.info(f"Using local temp path: {path}")
        return path

    def is_temp_table(self, name: str) -> bool:
        if any(
            pattern.match(name)
            for pattern in self.config._compiled_temporary_tables_pattern
        ):
            return True

        # This is also a temp table if
        #   1. this name would be allowed by the dataset patterns, and
        #   2. we have a list of discovered tables, and
        #   3. it's not in the discovered tables list
        if (
            self.filters.is_dataset_pattern_allowed(name, SnowflakeObjectDomain.TABLE)
            and self.discovered_tables
            and name not in self.discovered_tables
        ):
            return True

        return False

    def is_allowed_table(self, name: str) -> bool:
        if self.discovered_tables and name not in self.discovered_tables:
            return False

        return self.filters.is_dataset_pattern_allowed(
            name, SnowflakeObjectDomain.TABLE
        )

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        with self.report.users_fetch_timer:
            users = self.fetch_users()

        # TODO: Add some logic to check if the cached audit log is stale or not.
        audit_log_file = self.local_temp_path / "audit_log.sqlite"
        use_cached_audit_log = audit_log_file.exists()

        if self.config.local_temp_path is None:
            self._exit_stack.callback(lambda: audit_log_file.unlink(missing_ok=True))

        shared_connection = self._exit_stack.enter_context(
            ConnectionWrapper(audit_log_file)
        )
        queries: FileBackedList[
            Union[
                KnownLineageMapping,
                PreparsedQuery,
                TableRename,
                TableSwap,
                ObservedQuery,
                StoredProcCall,
            ]
        ] = self._exit_stack.enter_context(FileBackedList(shared_connection))

        if use_cached_audit_log:
            logger.info(f"Using cached audit log at {audit_log_file}")
        else:
            # Check if any query-based features are enabled before fetching
            needs_query_data = any(
                [
                    self.config.include_lineage,
                    self.config.include_queries,
                    self.config.include_usage_statistics,
                    self.config.include_query_usage_statistics,
                    self.config.include_operations,
                ]
            )

            if not needs_query_data:
                logger.info(
                    "All query-based features are disabled. Skipping expensive query log fetch."
                )
            else:
                logger.info(f"Fetching audit log into {audit_log_file}")

                with self.report.copy_history_fetch_timer:
                    for copy_entry in self.fetch_copy_history():
                        queries.append(copy_entry)

                with self.report.query_log_fetch_timer:
                    for entry in self.fetch_query_log(users):
                        queries.append(entry)

        stored_proc_tracker: StoredProcLineageTracker = self._exit_stack.enter_context(
            StoredProcLineageTracker(
                platform=self.identifiers.platform,
                shared_connection=shared_connection,
            )
        )
        self.report.stored_proc_lineage = stored_proc_tracker.report

        with self.report.audit_log_load_timer:
            for i, query in enumerate(queries):
                if i % 1000 == 0:
                    logger.info(f"Added {i} query log entries to SQL aggregator")

                if isinstance(query, StoredProcCall):
                    stored_proc_tracker.add_stored_proc_call(query)
                    continue

                if not (
                    isinstance(query, PreparsedQuery)
                    and stored_proc_tracker.add_related_query(query)
                ):
                    # Only add to aggregator if it's not part of a stored procedure.
                    self.aggregator.add(query)

            # Generate and add stored procedure lineage entries.
            for lineage_entry in stored_proc_tracker.build_merged_lineage_entries():
                # TODO: Make this the lowest priority lineage - so that it doesn't override other lineage entries.
                self.aggregator.add(lineage_entry)

        with self.report.aggregator_generate_timer:
            yield from auto_workunit(self.aggregator.gen_metadata())

        # Update the stateful ingestion state after successful extraction
        self._update_state()

    def fetch_users(self) -> UsersMapping:
        users: UsersMapping = dict()
        with self.structured_reporter.report_exc("Error fetching users from Snowflake"):
            logger.info("Fetching users from Snowflake")
            query = SnowflakeQuery.get_all_users()
            resp = self.connection.query(query)

            for row in resp:
                try:
                    users[row["NAME"]] = row["EMAIL"]
                    self.report.num_users += 1
                except Exception as e:
                    self.structured_reporter.warning(
                        "Error parsing user row",
                        context=f"{row}",
                        exc=e,
                    )
        return users

    def fetch_copy_history(self) -> Iterable[KnownLineageMapping]:
        # Derived from _populate_external_lineage_from_copy_history.

        query: str = SnowflakeQuery.copy_lineage_history(
            start_time_millis=int(self.start_time.timestamp() * 1000),
            end_time_millis=int(self.end_time.timestamp() * 1000),
            downstreams_deny_pattern=self.config.temporary_tables_pattern,
        )

        with self.structured_reporter.report_exc(
            "Error fetching copy history from Snowflake"
        ):
            logger.info("Fetching copy history from Snowflake")
            resp = self.connection.query(query)

            for row in resp:
                try:
                    result = (
                        SnowflakeLineageExtractor._process_external_lineage_result_row(
                            row,
                            discovered_tables=self.discovered_tables,
                            identifiers=self.identifiers,
                        )
                    )
                except Exception as e:
                    self.structured_reporter.warning(
                        "Error parsing copy history row",
                        context=f"{row}",
                        exc=e,
                    )
                else:
                    if result:
                        yield result

    def fetch_query_log(
        self, users: UsersMapping
    ) -> Iterable[
        Union[PreparsedQuery, TableRename, TableSwap, ObservedQuery, StoredProcCall]
    ]:
        query_log_query = QueryLogQueryBuilder(
            start_time=self.start_time,
            end_time=self.end_time,
            bucket_duration=self.config.window.bucket_duration,
            deny_usernames=self.config.pushdown_deny_usernames,
            allow_usernames=self.config.pushdown_allow_usernames,
            dedup_strategy=self.config.query_dedup_strategy,
            database_pattern=self.filters.filter_config.database_pattern
            if self.config.push_down_database_pattern_access_history
            else None,
            additional_database_names=self.config.additional_database_names_allowlist
            if self.config.push_down_database_pattern_access_history
            else None,
        ).build_enriched_query_log_query()

        with self.structured_reporter.report_exc(
            "Error fetching query log from Snowflake"
        ):
            logger.info("Fetching query log from Snowflake")
            resp = self.connection.query(query_log_query)

            for i, row in enumerate(resp):
                if i > 0 and i % 1000 == 0:
                    logger.info(f"Processed {i} query log rows so far")

                assert isinstance(row, dict)
                try:
                    yield from self._parse_audit_log_row(row, users)
                except Exception as e:
                    self.structured_reporter.warning(
                        "Error parsing query log row",
                        context=f"{row}",
                        exc=e,
                    )

    @classmethod
    def _has_temp_keyword(cls, query_text: str) -> bool:
        return (
            re.search(r"\bTEMP\b", query_text, re.IGNORECASE) is not None
            or re.search(r"\bTEMPORARY\b", query_text, re.IGNORECASE) is not None
        )

    @staticmethod
    def _classify_audit_log_objects(
        audit_log_objects: List[Dict[str, Any]],
    ) -> _AuditLogObjectClassification:
        """Scan an audit-log object list for name anomalies that force the stream bypass.

        Snowflake injects undocumented placeholder names (e.g. "$SYS_VIEW_<id>")
        into direct_objects_accessed and objects_modified for some stream-driven
        queries.  URNs built from those names are unusable, so any row containing
        them must fall back to sqlglot parsing via ObservedQuery (the bypass path
        introduced in PR #12318).

        We check only the *unqualified* (last dot-segment) name because the database
        and schema parts are normal for placeholder rows; the anomaly is always in
        the object name itself.

        Two separate canaries are tracked so downstream code can distinguish a known
        Snowflake behaviour ($SYS_VIEW_*) from a future undocumented shape ($-prefix
        that isn't $SYS_VIEW_) — the latter increments a drift counter so operators
        can detect if Snowflake introduces a new placeholder format.
        """
        has_placeholder = False
        n_missing_name = 0
        n_unknown_dollar = 0
        for obj in audit_log_objects:
            name = obj.get("objectName")
            if not name or not name.strip():
                n_missing_name += 1
                continue
            unqualified = name.rsplit(".", 1)[-1]
            if unqualified.startswith(_SYS_VIEW_PLACEHOLDER_PREFIX):
                has_placeholder = True
            elif unqualified.startswith("$"):
                n_unknown_dollar += 1
        return _AuditLogObjectClassification(
            has_placeholder=has_placeholder,
            n_missing_name=n_missing_name,
            n_unknown_dollar=n_unknown_dollar,
        )

    def _parse_audit_log_row(
        self, row: Dict[str, Any], users: UsersMapping
    ) -> Iterable[
        Union[TableRename, TableSwap, PreparsedQuery, ObservedQuery, StoredProcCall]
    ]:
        json_fields = {
            "DIRECT_OBJECTS_ACCESSED",
            "OBJECTS_MODIFIED",
            "OBJECT_MODIFIED_BY_DDL",
        }

        res = {}
        for key, value in row.items():
            if key in json_fields and value:
                value = json.loads(value)
            key = key.lower()
            res[key] = value

        timestamp: datetime = res["query_start_time"]
        timestamp = timestamp.astimezone(timezone.utc)

        # TODO need to map snowflake query types to ours
        query_text: str = res["query_text"]
        snowflake_query_type: str = res["query_type"]
        query_type: QueryType = SNOWFLAKE_QUERY_TYPE_MAPPING.get(
            snowflake_query_type, QueryType.UNKNOWN
        )

        direct_objects_accessed = res["direct_objects_accessed"] or []
        objects_modified = res["objects_modified"] or []
        object_modified_by_ddl = res["object_modified_by_ddl"]

        if (
            object_modified_by_ddl
            and not isinstance(object_modified_by_ddl, dict)
            and not objects_modified
        ):
            self.report.num_ddl_queries_dropped += 1
            if direct_objects_accessed:
                pass  # Fall through to normal query processing for usage tracking
            else:
                return

        if (
            isinstance(object_modified_by_ddl, dict)
            and object_modified_by_ddl
            and not objects_modified
        ):
            known_ddl_entry: Optional[Union[TableRename, TableSwap]] = None
            with self.structured_reporter.report_exc(
                "Error fetching ddl lineage from Snowflake"
            ):
                known_ddl_entry = self.parse_ddl_query(
                    query_text,
                    res["session_id"],
                    timestamp,
                    object_modified_by_ddl,
                    snowflake_query_type,
                )
            if known_ddl_entry:
                yield known_ddl_entry
                return
            elif direct_objects_accessed:
                # Unknown ddl relevant for usage. We want to continue execution here
                pass
            else:
                return

        user = CorpUserUrn(
            self.identifiers.get_user_identifier(
                res["user_name"], users.get(res["user_name"])
            )
        )
        extra_info = {
            "snowflake_query_id": res["query_id"],
            "snowflake_root_query_id": res["root_query_id"],
            "snowflake_query_type": res["query_type"],
            "snowflake_role_name": res["role_name"],
            "query_duration": res["query_duration"],
            "rows_inserted": res["rows_inserted"],
            "rows_updated": res["rows_updated"],
            "rows_deleted": res["rows_deleted"],
        }

        direct_cls = self._classify_audit_log_objects(direct_objects_accessed)
        modified_cls = self._classify_audit_log_objects(objects_modified)
        self.report.num_audit_rows_missing_object_name += (
            direct_cls.n_missing_name + modified_cls.n_missing_name
        )
        self.report.num_audit_rows_unknown_dollar_prefix += (
            direct_cls.n_unknown_dollar + modified_cls.n_unknown_dollar
        )

        has_stream_objects: bool = any(
            obj.get("objectDomain") == "Stream" for obj in direct_objects_accessed
        )
        has_corrupt_object_names: bool = (
            direct_cls.has_placeholder or modified_cls.has_placeholder
        )
        has_unknown_dollar_prefix: bool = (
            direct_cls.has_unknown_dollar_prefix
            or modified_cls.has_unknown_dollar_prefix
        )
        is_multi_target: bool = len(objects_modified) > 1
        is_create_view = query_type == QueryType.CREATE_VIEW
        is_create_temp_view = is_create_view and self._has_temp_keyword(query_text)

        stream_clean_multi_target: bool = (
            has_stream_objects
            and is_multi_target
            and not has_corrupt_object_names
            and not has_unknown_dollar_prefix
        )

        # Stream-bypass: Snowflake leaks $SYS_VIEW_<id> placeholder names into the
        # audit log for some stream-driven queries (undocumented; observed empirically).
        # URNs built from those names are unusable, so we fall back to sqlglot parsing
        # via ObservedQuery.  Single-target stream queries also stay on this path
        # because PR #15391's per-downstream logic only helps multi-target INSERT ALL;
        # routing a single-target query through it would add no value.
        # CREATE TEMPORARY VIEW takes the bypass regardless of stream presence because
        # sqlglot resolves temp-view references correctly without audit-log metadata.
        if (
            has_stream_objects and not stream_clean_multi_target
        ) or is_create_temp_view:
            if has_stream_objects:
                self.report.num_stream_queries_observed += 1
                logger.debug(
                    "Snowflake stream-bypass fired for query_id=%s "
                    "(corrupt=%s, unknown_dollar=%s, multi_target=%s)",
                    res["query_id"],
                    has_corrupt_object_names,
                    has_unknown_dollar_prefix,
                    is_multi_target,
                )
            elif is_create_temp_view:
                self.report.num_create_temp_view_queries_observed += 1

            yield ObservedQuery(
                query=query_text,
                session_id=res["session_id"],
                timestamp=timestamp,
                user=user,
                default_db=res["default_db"],
                default_schema=res["default_schema"],
                query_hash=get_query_fingerprint(
                    query_text, self.identifiers.platform, fast=True
                ),
                extra_info=extra_info,
            )
            return

        # Reaching here with has_stream_objects=True means stream_clean_multi_target
        # is True (the bypass above returned early for all other stream cases).
        if stream_clean_multi_target:
            self.report.num_stream_queries_clean_fast_path += 1

        if snowflake_query_type == "CALL" and res["root_query_id"] is None:
            yield StoredProcCall(
                # This is the top-level query ID that other entries will reference.
                snowflake_root_query_id=res["query_id"],
                query_text=query_text,
                timestamp=timestamp,
                user=user,
                default_db=res["default_db"],
                default_schema=res["default_schema"],
            )
            return

        upstreams = []
        column_usage = {}

        for obj in direct_objects_accessed:
            obj_name = obj.get("objectName")
            if not obj_name or not obj_name.strip():
                logger.debug(
                    "Skipping direct_objects_accessed entry with missing objectName "
                    "for query_id=%s",
                    res["query_id"],
                )
                continue
            dataset = self.identifiers.gen_dataset_urn(
                self.identifiers.get_dataset_identifier_from_qualified_name(obj_name)
            )

            columns = set()
            for modified_column in obj.get("columns") or []:
                column_name = modified_column.get("columnName")
                # An empty or missing column name in the audit log would cause an error when creating column URNs.
                # To avoid this and still extract lineage, the raw query text is parsed as a fallback.
                if not column_name or not column_name.strip():
                    query_id = res["query_id"]
                    self.report.num_queries_with_empty_column_name += 1
                    self.report.queries_with_empty_column_name.append(query_id)
                    logger.info(f"Query {query_id} has empty column name in audit log.")

                    yield ObservedQuery(
                        query=query_text,
                        session_id=res["session_id"],
                        timestamp=timestamp,
                        user=user,
                        default_db=res["default_db"],
                        default_schema=res["default_schema"],
                        query_hash=get_query_fingerprint(
                            query_text, self.identifiers.platform, fast=True
                        ),
                        extra_info=extra_info,
                    )
                    return
                columns.add(self.identifiers.snowflake_identifier(column_name))

            upstreams.append(dataset)
            column_usage[dataset] = columns

        if not upstreams and direct_objects_accessed:
            # Every upstream entry was dropped (missing objectName). Falling back
            # to sqlglot is safer than emitting a PreparsedQuery with upstreams=[]
            # — same-fingerprint rows would overwrite previously-correct lineage
            # in SqlParsingAggregator's query map.
            self.structured_reporter.warning(
                "All direct_objects_accessed entries had missing objectName; "
                "falling back to SQL-based lineage",
                context=f"query_id={res['query_id']}, "
                f"num_objects={len(direct_objects_accessed)}, "
                f"n_missing={direct_cls.n_missing_name}",
            )
            yield ObservedQuery(
                query=query_text,
                session_id=res["session_id"],
                timestamp=timestamp,
                user=user,
                default_db=res["default_db"],
                default_schema=res["default_schema"],
                usage_multiplier=res["query_count"],
                query_hash=get_query_fingerprint(
                    query_text, self.identifiers.platform, fast=True
                ),
                extra_info=extra_info,
            )
            return

        targets = self._build_downstream_targets(
            objects_modified,
            res["query_secondary_fingerprint"],
            is_multi_target,
            query_id=res["query_id"],
        )

        if objects_modified and not targets:
            self.structured_reporter.warning(
                "All downstream objects failed; falling back to SQL-based lineage",
                context=f"query_id={res['query_id']}, "
                f"num_objects={len(objects_modified)}",
            )
            yield ObservedQuery(
                query=query_text,
                session_id=res["session_id"],
                timestamp=timestamp,
                user=user,
                default_db=res["default_db"],
                default_schema=res["default_schema"],
                usage_multiplier=res["query_count"],
                query_hash=get_query_fingerprint(
                    query_text, self.identifiers.platform, fast=True
                ),
                extra_info=extra_info,
            )
            return

        for i, target in enumerate(targets):
            # Only the first entry carries query_count; subsequent entries get 0
            # to avoid double-counting upstream reads when multiple targets are yielded.
            entry_query_count = res["query_count"] if i == 0 else 0

            yield PreparsedQuery(
                # Use DataHub's own fingerprint instead of Snowflake's opaque
                # QUERY_SECONDARY_FINGERPRINT. fast=True uses regex-based
                # normalization instead of full sqlglot parsing.
                query_id=get_query_fingerprint(
                    query_text,
                    self.identifiers.platform,
                    fast=True,
                    secondary_id=target.secondary_id,
                ),
                query_text=query_text,
                upstreams=list(upstreams),
                downstream=target.downstream,
                column_lineage=target.column_lineage,
                column_usage=dict(column_usage),
                inferred_schema=None,
                confidence_score=1.0,
                query_count=entry_query_count,
                user=user,
                timestamp=timestamp,
                session_id=res["session_id"],
                query_type=query_type,
                extra_info=extra_info,
            )

    def _build_downstream_targets(
        self,
        objects_modified: List[Dict[str, Any]],
        query_secondary_fingerprint: Optional[str],
        is_multi_target: bool,
        query_id: str,
    ) -> List[_DownstreamTarget]:
        """Build one _DownstreamTarget per modified object in an audit log row.

        For multi-target INSERTs (INSERT ALL / INSERT FIRST) this returns N entries;
        for normal queries it returns one. Each object is isolated so a bad entry
        doesn't discard siblings. For queries with no objects_modified (e.g. SELECT
        statements), returns a single entry with downstream=None to preserve
        upstream usage tracking. If all objects fail to process (e.g. malformed
        audit log entries), returns an empty list; the caller falls back to
        SQL-based lineage.
        """
        targets: List[_DownstreamTarget] = []

        if not objects_modified:
            targets.append(
                _DownstreamTarget(
                    downstream=None,
                    column_lineage=None,
                    secondary_id=query_secondary_fingerprint,
                )
            )
        else:
            for obj in objects_modified:
                try:
                    obj_downstream = self.identifiers.gen_dataset_urn(
                        self.identifiers.get_dataset_identifier_from_qualified_name(
                            obj["objectName"]
                        )
                    )
                    obj_column_lineage: List[ColumnLineageInfo] = []
                    for modified_column in obj["columns"]:
                        obj_column_lineage.append(
                            ColumnLineageInfo(
                                downstream=DownstreamColumnRef(
                                    table=obj_downstream,
                                    column=self.identifiers.snowflake_identifier(
                                        modified_column["columnName"]
                                    ),
                                ),
                                upstreams=[
                                    ColumnRef(
                                        table=self.identifiers.gen_dataset_urn(
                                            self.identifiers.get_dataset_identifier_from_qualified_name(
                                                upstream["objectName"]
                                            )
                                        ),
                                        column=self.identifiers.snowflake_identifier(
                                            upstream["columnName"]
                                        ),
                                    )
                                    for upstream in modified_column["directSources"]
                                    if upstream["objectDomain"]
                                    in SnowflakeQuery.ACCESS_HISTORY_TABLE_VIEW_DOMAINS
                                ],
                            )
                        )

                    # Multi-target INSERTs need a unique query_id per downstream table
                    # so SqlParsingAggregator treats each as a distinct lineage entry.
                    secondary_id = query_secondary_fingerprint
                    if is_multi_target:
                        base = secondary_id or ""
                        secondary_id = f"{base}_{obj['objectName']}"

                    targets.append(
                        _DownstreamTarget(
                            downstream=obj_downstream,
                            column_lineage=obj_column_lineage,
                            secondary_id=secondary_id,
                        )
                    )
                except (KeyError, TypeError, ValueError) as e:
                    self.structured_reporter.warning(
                        "Failed to process downstream object",
                        context=f"query_id={query_id}, "
                        f"object={obj.get('objectName') or '<unknown>'}",
                        exc=e,
                    )

        return targets

    @staticmethod
    def _get_ddl_property(properties: dict, key: str) -> str:
        """Extract a property value from object_modified_by_ddl properties,
        handling both pre-2026_01 BCR ({"key": {"value": "..."}}) and
        post-2026_01 BCR ({"key": "..."}) formats.

        Callers must ensure properties[key] exists and is truthy before calling."""
        val = properties[key]
        if isinstance(val, dict):
            return val["value"]
        return val

    def parse_ddl_query(
        self,
        query: str,
        session_id: str,
        timestamp: datetime,
        object_modified_by_ddl: object,
        query_type: str,
    ) -> Optional[Union[TableRename, TableSwap]]:
        if not isinstance(object_modified_by_ddl, dict):
            self.report.num_ddl_queries_dropped += 1
            return None
        if (
            object_modified_by_ddl["operationType"] == "ALTER"
            and query_type == "RENAME_TABLE"
            and object_modified_by_ddl["properties"].get("objectName")
        ):
            original_un = self.identifiers.gen_dataset_urn(
                self.identifiers.get_dataset_identifier_from_qualified_name(
                    object_modified_by_ddl["objectName"]
                )
            )

            new_urn = self.identifiers.gen_dataset_urn(
                self.identifiers.get_dataset_identifier_from_qualified_name(
                    self._get_ddl_property(
                        object_modified_by_ddl["properties"], "objectName"
                    )
                )
            )
            return TableRename(original_un, new_urn, query, session_id, timestamp)
        elif object_modified_by_ddl[
            "operationType"
        ] == "ALTER" and object_modified_by_ddl["properties"].get("swapTargetName"):
            urn1 = self.identifiers.gen_dataset_urn(
                self.identifiers.get_dataset_identifier_from_qualified_name(
                    object_modified_by_ddl["objectName"]
                )
            )

            urn2 = self.identifiers.gen_dataset_urn(
                self.identifiers.get_dataset_identifier_from_qualified_name(
                    self._get_ddl_property(
                        object_modified_by_ddl["properties"], "swapTargetName"
                    )
                )
            )

            return TableSwap(urn1, urn2, query, session_id, timestamp)
        else:
            self.report.num_ddl_queries_dropped += 1
            return None

    def close(self) -> None:
        self._exit_stack.close()


@support_status(SupportStatus.CERTIFIED)
@config_class(SnowflakeQueriesSourceConfig)
class SnowflakeQueriesSource(Source):
    def __init__(self, ctx: PipelineContext, config: SnowflakeQueriesSourceConfig):
        self.ctx = ctx
        self.config = config
        self.report = SnowflakeQueriesSourceReport()

        self.filters = SnowflakeFilter(
            filter_config=self.config,
            structured_reporter=self.report,
        )
        self.identifiers = SnowflakeIdentifierBuilder(
            identifier_config=self.config,
            structured_reporter=self.report,
        )

        self.connection = self.config.connection.get_connection()

        self.queries_extractor = SnowflakeQueriesExtractor(
            connection=self.connection,
            config=self.config,
            structured_report=self.report,
            filters=self.filters,
            identifiers=self.identifiers,
            graph=self.ctx.graph,
        )
        self.report.queries_extractor = self.queries_extractor.report

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> Self:
        config = SnowflakeQueriesSourceConfig.model_validate(config_dict)
        return cls(ctx, config)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        self.report.window = self.config.window

        # TODO: Disable auto status processor?
        return self.queries_extractor.get_workunits_internal()

    def get_report(self) -> SnowflakeQueriesSourceReport:
        return self.report

    def close(self) -> None:
        self.connection.close()
        self.queries_extractor.close()
        super().close()


class QueryLogQueryBuilder:
    def __init__(
        self,
        start_time: datetime,
        end_time: datetime,
        bucket_duration: BucketDuration,
        deny_usernames: Optional[List[str]] = None,
        allow_usernames: Optional[List[str]] = None,
        max_tables_per_query: int = 20,
        dedup_strategy: QueryDedupStrategyType = QueryDedupStrategyType.STANDARD,
        database_pattern: Optional[AllowDenyPattern] = None,
        additional_database_names: Optional[List[str]] = None,
    ):
        self.start_time = start_time
        self.end_time = end_time
        self.start_time_millis = int(start_time.timestamp() * 1000)
        self.end_time_millis = int(end_time.timestamp() * 1000)
        self.max_tables_per_query = max_tables_per_query
        self.dedup_strategy = dedup_strategy

        self.users_filter = self._build_user_filter(deny_usernames, allow_usernames)

        self.access_history_database_filter = (
            self._build_access_history_database_filter_condition(
                database_pattern, additional_database_names
            )
        )

        self.time_bucket_size = bucket_duration.value
        assert self.time_bucket_size in ("HOUR", "DAY", "MONTH")

    def _build_user_filter(
        self,
        deny_usernames: Optional[List[str]] = None,
        allow_usernames: Optional[List[str]] = None,
    ) -> str:
        """
        Build user filter SQL condition based on deny and allow username patterns.

        Args:
            deny_usernames: List of username patterns to exclude (SQL LIKE patterns)
            allow_usernames: List of username patterns to include (SQL LIKE patterns)

        Returns:
            SQL WHERE condition string for filtering users
        """
        user_filters = []

        if deny_usernames:
            deny_conditions = []
            for pattern in deny_usernames:
                # Escape single quotes for SQL safety
                escaped_pattern = pattern.replace("'", "''")
                deny_conditions.append(f"user_name NOT ILIKE '{escaped_pattern}'")
            if deny_conditions:
                user_filters.append(f"({' AND '.join(deny_conditions)})")

        if allow_usernames:
            allow_conditions = []
            for pattern in allow_usernames:
                # Escape single quotes for SQL safety
                escaped_pattern = pattern.replace("'", "''")
                allow_conditions.append(f"user_name ILIKE '{escaped_pattern}'")
            if allow_conditions:
                user_filters.append(f"({' OR '.join(allow_conditions)})")

        return " AND ".join(user_filters) if user_filters else "TRUE"

    def _build_access_history_database_filter_condition(
        self,
        database_pattern: Optional[AllowDenyPattern],
        additional_database_names: Optional[List[str]] = None,
    ) -> str:
        """
        Build a SQL WHERE condition for database filtering in access_history based on AllowDenyPattern.

        IMPORTANT: This function handles the fundamental difference between DML and DDL operations in Snowflake's
        access_history table:

        - DML Operations (SELECT, INSERT, UPDATE, DELETE, etc.): Store accessed/modified objects in the
          `direct_objects_accessed` and `objects_modified` arrays
        - DDL Operations (CREATE, ALTER, DROP, RENAME, etc.): Store modified objects in the
          `object_modified_by_ddl` field (single object, not an array)

        Without checking `object_modified_by_ddl`, DDL operations like "ALTER TABLE person_info RENAME TO person_info_final"
        would be incorrectly filtered out because they don't populate the DML arrays, causing missing lineage
        and operational metadata.

        Filtering Logic:
        A query is included if it matches:
        - Any database name in additional_database_names (exact match), OR
        - Any database pattern in database_pattern.allow AND NOT any pattern in database_pattern.deny

        Args:
            database_pattern: The AllowDenyPattern configuration for database filtering
            additional_database_names: Additional database names to always include (no pattern matching)

        Returns:
            A SQL WHERE condition string, or "TRUE" if no filtering should be applied
        """
        if not database_pattern and not additional_database_names:
            return "TRUE"

        # Build the database filter conditions
        # Logic: Allow if (matches additional_database_names_allowlist) OR (matches database_pattern.allow AND NOT matches database_pattern.deny)
        # Note: Using UPPER() + RLIKE for case-insensitive matching is more performant than REGEXP_LIKE with 'i' flag

        # Build additional database names condition (exact matches) - these always get included
        additional_db_condition = None
        if additional_database_names:
            additional_db_conditions = []
            for db_name in additional_database_names:
                # Escape single quotes
                escaped_db_name = db_name.replace("'", "''")
                additional_db_conditions.append(
                    f"SPLIT_PART(UPPER(o:objectName), '.', 1) = '{escaped_db_name.upper()}'"
                )
            if additional_db_conditions:
                additional_db_condition = " OR ".join(additional_db_conditions)

        # Build database pattern condition (allow AND NOT deny)
        database_pattern_condition = None
        if database_pattern:
            allow_patterns = database_pattern.allow
            deny_patterns = database_pattern.deny

            pattern_parts = []

            # Add allow patterns (if not the default "allow all")
            if allow_patterns and allow_patterns != [".*"]:
                allow_conditions = []
                for pattern in allow_patterns:
                    # Escape single quotes that might be present in the regex pattern
                    escaped_pattern = pattern.replace("'", "''")
                    allow_conditions.append(
                        f"SPLIT_PART(UPPER(o:objectName), '.', 1) RLIKE '{escaped_pattern}'"
                    )
                if allow_conditions:
                    pattern_parts.append(
                        allow_conditions[0]
                        if len(allow_conditions) == 1
                        else f"({' OR '.join(allow_conditions)})"
                    )

            # Add deny patterns
            if deny_patterns:
                deny_conditions = []
                for pattern in deny_patterns:
                    # Escape single quotes that might be present in the regex pattern
                    escaped_pattern = pattern.replace("'", "''")
                    deny_conditions.append(
                        f"SPLIT_PART(UPPER(o:objectName), '.', 1) NOT RLIKE '{escaped_pattern}'"
                    )
                if deny_conditions:
                    pattern_parts.append(
                        deny_conditions[0]
                        if len(deny_conditions) == 1
                        else f"({' AND '.join(deny_conditions)})"
                    )

            if pattern_parts:
                database_pattern_condition = " AND ".join(pattern_parts)

        # Combine conditions: additional_database_names OR database_pattern
        filter_conditions = []
        if additional_db_condition:
            filter_conditions.append(
                f"({additional_db_condition})"
                if len(additional_db_condition.split(" OR ")) > 1
                else additional_db_condition
            )
        if database_pattern_condition:
            filter_conditions.append(
                f"({database_pattern_condition})"
                if len(database_pattern_condition.split(" AND ")) > 1
                else database_pattern_condition
            )

        if filter_conditions:
            database_filter_condition = (
                filter_conditions[0]
                if len(filter_conditions) == 1
                else " OR ".join(filter_conditions)
            )

            # Build a condition that checks if any objects in the arrays match the database pattern
            # This implements "at least one" matching behavior: queries are allowed if they touch
            # at least one database that matches the pattern, even if they also touch other databases
            # Use ARRAY_SIZE with FILTER which is more compatible with Snowflake
            direct_objects_condition = f"ARRAY_SIZE(FILTER(direct_objects_accessed, o -> {database_filter_condition})) > 0"
            objects_modified_condition = f"ARRAY_SIZE(FILTER(objects_modified, o -> {database_filter_condition})) > 0"

            # CRITICAL: Handle DDL operations by checking object_modified_by_ddl field
            # DDL operations like ALTER TABLE RENAME store their data here instead of in the arrays
            # We need to adapt the filter condition for a single object rather than an array
            ddl_filter_condition = database_filter_condition.replace(
                "o:objectName", "object_modified_by_ddl:objectName"
            )
            object_modified_by_ddl_condition = f"({ddl_filter_condition})"

            return f"({direct_objects_condition} OR {objects_modified_condition} OR {object_modified_by_ddl_condition})"
        else:
            return "TRUE"

    def _query_fingerprinted_queries(self):
        if self.dedup_strategy == QueryDedupStrategyType.STANDARD:
            secondary_fingerprint_sql = """
    CASE 
        WHEN CONTAINS(query_history.query_text, '-- Hex query metadata:')
        -- Extract project id and hash it
        THEN CAST(HASH(
            REGEXP_SUBSTR(query_history.query_text, '"project_id"\\\\s*:\\\\s*"([^"]+)"', 1, 1, 'e', 1),
            REGEXP_SUBSTR(query_history.query_text, '"context"\\\\s*:\\\\s*"([^"]+)"', 1, 1, 'e', 1)
        ) AS VARCHAR)
        ELSE NULL 
    END"""
        elif self.dedup_strategy == QueryDedupStrategyType.NONE:
            secondary_fingerprint_sql = "NULL"
        else:
            raise NotImplementedError(
                f"Strategy {self.dedup_strategy} is not implemented by the QueryLogQueryBuilder"
            )
        return f"""
SELECT *,
    -- TODO: Generate better fingerprints for each query by pushing down regex logic.
    query_history.query_parameterized_hash as query_fingerprint,
    -- Optional and additional hash to be used for query deduplication and final query identity
    {secondary_fingerprint_sql} as query_secondary_fingerprint
FROM
    snowflake.account_usage.query_history
WHERE
    query_history.start_time >= to_timestamp_ltz({self.start_time_millis}, 3) -- {self.start_time.isoformat()}
    AND query_history.start_time < to_timestamp_ltz({self.end_time_millis}, 3) -- {self.end_time.isoformat()}
    AND execution_status = 'SUCCESS'
    AND {self.users_filter}"""

    def _query_deduplicated_queries(self):
        if self.dedup_strategy == QueryDedupStrategyType.STANDARD:
            return f"""
SELECT
    *,
    DATE_TRUNC(
        {self.time_bucket_size},
        CONVERT_TIMEZONE('UTC', start_time)
    ) AS bucket_start_time,
    COUNT(*) OVER (PARTITION BY bucket_start_time, query_fingerprint, query_secondary_fingerprint) AS query_count,
FROM
    fingerprinted_queries
QUALIFY
    ROW_NUMBER() OVER (PARTITION BY bucket_start_time, query_fingerprint, query_secondary_fingerprint ORDER BY start_time DESC) = 1"""
        elif self.dedup_strategy == QueryDedupStrategyType.NONE:
            return f"""
SELECT
    *,
    DATE_TRUNC(
        {self.time_bucket_size},
        CONVERT_TIMEZONE('UTC', start_time)
    ) AS bucket_start_time,
    1 AS query_count,
FROM
            fingerprinted_queries"""
        else:
            raise NotImplementedError(
                f"Strategy {self.dedup_strategy} is not implemented by the QueryLogQueryBuilder"
            )

    def build_enriched_query_log_query(self) -> str:
        return f"""\
WITH
fingerprinted_queries as (
{self._query_fingerprinted_queries()}
)
, deduplicated_queries as (
{self._query_deduplicated_queries()}
)
, raw_access_history AS (
    SELECT
        query_id,
        root_query_id,
        query_start_time,
        user_name,
        direct_objects_accessed,
        objects_modified,
        object_modified_by_ddl
    FROM
        snowflake.account_usage.access_history
    WHERE
        query_start_time >= to_timestamp_ltz({self.start_time_millis}, 3) -- {self.start_time.isoformat()}
        AND query_start_time < to_timestamp_ltz({self.end_time_millis}, 3) -- {self.end_time.isoformat()}
        AND {self.users_filter}
        AND query_id IN (
            SELECT query_id FROM deduplicated_queries
        )
        AND {self.access_history_database_filter}
)
, filtered_access_history AS (
    -- TODO: Add table filter clause.
    SELECT
        query_id,
        root_query_id,
        query_start_time,
        ARRAY_SLICE(
            FILTER(direct_objects_accessed, o -> o:objectDomain IN {SnowflakeQuery.ACCESS_HISTORY_TABLE_VIEW_DOMAINS_FILTER}),
            0, {self.max_tables_per_query}
        ) as direct_objects_accessed,
        -- TODO: Drop the columns.baseSources subfield.
        FILTER(objects_modified, o -> o:objectDomain IN {SnowflakeQuery.ACCESS_HISTORY_TABLE_VIEW_DOMAINS_FILTER}) as objects_modified,
        case when object_modified_by_ddl:objectDomain IN {SnowflakeQuery.ACCESS_HISTORY_TABLE_VIEW_DOMAINS_FILTER} then object_modified_by_ddl else null end as object_modified_by_ddl
    FROM raw_access_history
    WHERE ( array_size(direct_objects_accessed) > 0 or array_size(objects_modified) > 0 or object_modified_by_ddl is not null )
)
, query_access_history AS (
    SELECT
        q.bucket_start_time,
        q.query_id,
        q.query_fingerprint,
        q.query_secondary_fingerprint,
        q.query_count,
        q.session_id AS "SESSION_ID",
        q.start_time AS "QUERY_START_TIME",
        q.total_elapsed_time AS "QUERY_DURATION",
        q.query_text AS "QUERY_TEXT",
        q.query_type AS "QUERY_TYPE",
        q.database_name as "DEFAULT_DB",
        q.schema_name as "DEFAULT_SCHEMA",
        q.rows_inserted AS "ROWS_INSERTED",
        q.rows_updated AS "ROWS_UPDATED",
        q.rows_deleted AS "ROWS_DELETED",
        q.user_name AS "USER_NAME",
        q.role_name AS "ROLE_NAME",
        a.root_query_id,
        a.direct_objects_accessed,
        a.objects_modified,
        a.object_modified_by_ddl
    FROM deduplicated_queries q
    JOIN filtered_access_history a USING (query_id)
)
SELECT * FROM query_access_history
-- Our query aggregator expects the queries to be added in chronological order.
-- It's easier for us to push down the sorting to Snowflake/SQL instead of doing it in Python.
ORDER BY QUERY_START_TIME ASC
"""


SNOWFLAKE_QUERY_TYPE_MAPPING = {
    "INSERT": QueryType.INSERT,
    "UPDATE": QueryType.UPDATE,
    "DELETE": QueryType.DELETE,
    "CREATE": QueryType.CREATE_OTHER,
    "CREATE_TABLE": QueryType.CREATE_DDL,
    "CREATE_VIEW": QueryType.CREATE_VIEW,
    "CREATE_TABLE_AS_SELECT": QueryType.CREATE_TABLE_AS_SELECT,
    "MERGE": QueryType.MERGE,
    "COPY": QueryType.INSERT,
    "TRUNCATE_TABLE": QueryType.UNKNOWN,
}



import logging
import re
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import requests
from pydantic import ValidationError

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import ContainerKey
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    MetadataWorkUnitProcessor,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.informatica.client import InformaticaClient
from datahub.ingestion.source.informatica.config import (
    CONNECTION_TYPE_MAP,
    InformaticaSourceConfig,
)
from datahub.ingestion.source.informatica.models import (
    ExportJobState,
    IdmcConnection,
    IdmcMapping,
    IdmcMappingTask,
    IdmcObject,
    IdmcObjectType,
    InformaticaApiError,
    InformaticaLoginError,
    InformaticaSourceReport,
    LineageTable,
    MappingLineageInfo,
    TaskflowStep,
    V2Id,
    V3Guid,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.schema_classes import (
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    DataJobInputOutputClass,
    DatasetLineageTypeClass,
    OwnershipTypeClass,
    StatusClass,
    UpstreamClass,
    UpstreamLineageClass,
)
from datahub.metadata.urns import CorpUserUrn, DatasetUrn, TagUrn
from datahub.sdk.container import Container
from datahub.sdk.dataflow import DataFlow
from datahub.sdk.datajob import DataJob
from datahub.sdk.entity import Entity

logger = logging.getLogger(__name__)

PLATFORM = "informatica"
ORPHAN_PROJECT_SENTINEL = "__root__"
# Shared between extraction and lineage phases so DataJob URNs match.
MAPPING_JOB_ID = "transform"

_UNSAFE_FLOW_ID_CHARS = re.compile(r"[^\w-]+")

T = TypeVar("T")


def _safe_flow_id(name: str, fallback: str) -> str:
    """URN-safe flow_id from a user-facing IDMC name.

    DataHub's Navigate tree renders DataFlow nodes by the URN flow_id
    segment (not ``dataFlowInfo.name``), and visually splits on ``.``
    and ``/`` — so we sanitize those out. Falls back to ``fallback``
    when the result is empty so we never emit an MCP with an empty URN.
    """
    if not name:
        return fallback
    sanitized = _UNSAFE_FLOW_ID_CHARS.sub("_", name).strip("_")
    return sanitized or fallback


@dataclass
class OrchestrateState:
    """Per-orchestrate state accumulated across the lineage phase.

    ``output_datasets`` fills from the last MT's outputs (via the
    ``_orchestrate_by_last_mt`` reverse index) before emission.
    """

    last_mt_urn: str
    mt_urns_in_order: List[str]
    output_datasets: Set[str] = field(default_factory=set)


@dataclass
class TaskflowStepWalkResult:
    step_order: List[str]
    mt_urns_in_order: List[str]
    step_summary_parts: List[str]


class InformaticaProjectKey(ContainerKey):
    project_name: str


class InformaticaFolderKey(InformaticaProjectKey):
    # Inherit from ProjectKey so ContainerKey.parent_key() walks
    # Folder → Project automatically for BrowsePathsV2 assembly.
    folder_name: str


@platform_name("Informatica")
@config_class(InformaticaSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.CONTAINERS, "Projects and folders as containers")
@capability(SourceCapability.LINEAGE_COARSE, "Table-level lineage via v3 Export API")
@capability(SourceCapability.DELETION_DETECTION, "Via stateful ingestion")
@capability(SourceCapability.OWNERSHIP, "From IDMC object createdBy/updatedBy")
@capability(SourceCapability.TAGS, "IDMC object tags emitted as DataHub GlobalTags")
@capability(SourceCapability.TEST_CONNECTION, "Enabled by default")
class InformaticaSource(StatefulIngestionSourceBase, TestableSource):
    """DataHub ingestion source for Informatica Cloud (IDMC).

    Emits IDMC Projects and Folders as Containers; Taskflows and Mapping
    Tasks as DataFlows (each MT with an inner ``transform`` DataJob that
    carries table-level lineage resolved from the referenced Mapping). Plain
    Mappings and Mapplets are intentionally not emitted — only runnable MTs
    surface in DataHub.
    """

    config: InformaticaSourceConfig
    report: InformaticaSourceReport
    platform: str = PLATFORM

    def __init__(self, config: InformaticaSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.config = config
        self.report = InformaticaSourceReport()
        self.client = InformaticaClient(config, self.report)
        # Lineage lookup by federated_id; config overrides by id.
        self._connections_by_fed_id: Dict[str, IdmcConnection] = {}
        self._connections_by_id: Dict[str, IdmcConnection] = {}
        # Two indexes into v2 /api/v2/mapping — MT→Mapping refs use v2 ids,
        # export batches use v3 GUIDs.
        self._v2_mappings_by_guid: Dict[V3Guid, IdmcMapping] = {}
        self._v2_mappings_by_v2_id: Dict[V2Id, IdmcMapping] = {}
        self._mapping_ids: List[V3Guid] = []
        # v3 GUID → MT DataJob URNs that run that mapping (fan-out).
        self._mapping_v3_to_mt_job_urns: Dict[V3Guid, List[str]] = {}
        self._mt_v2_id_to_job_urn: Dict[V2Id, str] = {}
        # MT DataJob URN → its step-order predecessors in any Taskflow.
        # Ordered + deduplicated so the same predecessor isn't added twice.
        self._mt_predecessors: Dict[str, List[str]] = {}
        # Orchestrate sits at the end of a Taskflow's chain:
        # ``input_ds → MT1 → … → MTn → orchestrate → out_ds``.
        self._orchestrate_state: Dict[str, OrchestrateState] = {}
        # Reverse index so lineage emission bubbles the last MT's outputs
        # up to any Taskflow's orchestrate.
        self._orchestrate_by_last_mt: Dict[str, List[str]] = {}
        # Containers we've emitted — filtered segments fall back to
        # plain-name browse entries instead of dangling URNs.
        self._emitted_project_names: Set[str] = set()
        self._emitted_folder_keys: Set[Tuple[str, str]] = set()
        # External datasets (Snowflake, Oracle, etc.) referenced by
        # lineage need a minimal Status aspect to register as existing
        # entities; otherwise ``searchAcrossLineage`` filters them out.
        self._stubbed_external_dataset_urns: Set[str] = set()

    @classmethod
    def create(cls, config_dict: Dict, ctx: PipelineContext) -> "InformaticaSource":
        config = InformaticaSourceConfig.model_validate(config_dict)
        return cls(config, ctx)

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        """Validate login credentials against IDMC before a full ingestion run."""
        try:
            config = InformaticaSourceConfig.model_validate(config_dict)
        except (ValidationError, ValueError, TypeError) as e:
            return TestConnectionReport(
                internal_failure=True,
                internal_failure_reason=f"Failed to parse config: {e}",
            )
        client = InformaticaClient(config, InformaticaSourceReport())
        try:
            client.login()
            return TestConnectionReport(
                basic_connectivity=CapabilityReport(capable=True)
            )
        except InformaticaLoginError as e:
            return TestConnectionReport(
                basic_connectivity=CapabilityReport(
                    capable=False,
                    failure_reason=str(e),
                )
            )
        except Exception as e:
            return TestConnectionReport(
                internal_failure=True,
                internal_failure_reason=f"Unexpected error during login: {e}",
            )
        finally:
            client.close()

    def get_report(self) -> InformaticaSourceReport:
        return self.report

    def close(self) -> None:
        self.client.close()
        super().close()

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(
        self,
    ) -> Iterable[Union[MetadataWorkUnit, Entity]]:
        try:
            self.client.login()
        except InformaticaLoginError as e:
            self.report.failure(
                title="IDMC login failed",
                message="Cannot authenticate to Informatica Cloud; aborting ingestion. Check login_url, username, and password.",
                context=self.config.login_url,
                exc=e,
            )
            return
        yield from self._extract_containers()
        # MTs emit before Taskflows so step resolution sees real MT URNs.
        yield from self._extract_mappings_and_tasks()
        yield from self._extract_taskflows()
        self._summarize_taskflow_steps()
        if self.config.extract_lineage:
            yield from self._extract_lineage()

    def _summarize_taskflow_steps(self) -> None:
        """Surface step-DAG coverage as a report warning when incomplete
        so missing ``Asset - export`` privilege is actionable from the
        ingestion report JSON without scanning logs.
        """
        total = self.report.taskflows_scanned
        with_steps = self.report.taskflows_with_steps
        if total == 0 or with_steps == total:
            return
        missing = total - with_steps
        self.report.warning(
            title="Taskflow step DAG missing for some Taskflows",
            message=(
                f"Resolved {with_steps}/{total} Taskflow step DAGs; "
                f"{missing} Taskflow(s) emitted as DataFlow-only. "
                "Common cause: service account lacks 'Asset - export' "
                "privilege. Grant it to enable MT chaining via inputDatajobs."
            ),
            context=f"taskflows_scanned={total}, taskflows_with_steps={with_steps}",
        )

    def _extract_containers(self) -> Iterable[Entity]:
        # Projects emit before Folders so folder parent_container edges
        # can check ``_emitted_project_names`` and skip orphans.
        yield from self._emit_projects()
        yield from self._emit_folders()

    def _emit_projects(self) -> Iterable[Entity]:
        projects = self._safe_list(
            lambda: self._iter_with_tags("Project"),
            title="Failed to list IDMC projects",
            message="Container hierarchy will be incomplete.",
            context="/public/core/v3/objects?type=Project",
        )
        for project in projects:
            if not self.config.project_pattern.allowed(project.name):
                self.report.projects_filtered += 1
                self.report.report_filtered(
                    "pattern", "Project", f"name={project.name!r} path={project.path}"
                )
                continue
            self.report.projects_scanned += 1
            self._emitted_project_names.add(project.name)
            logger.debug(
                "Emitting Project container: name=%s path=%s id=%s",
                project.name,
                project.path,
                project.id,
            )
            yield self._make_project_container(project)

    def _emit_folders(self) -> Iterable[Entity]:
        folders = self._safe_list(
            lambda: self._iter_with_tags("Folder"),
            title="Failed to list IDMC folders",
            message="Container hierarchy will be incomplete.",
            context="/public/core/v3/objects?type=Folder",
        )
        for folder in folders:
            try:
                if self._is_project_masquerading_as_folder(folder):
                    self.report.report_filtered(
                        "project-level",
                        "Folder",
                        f"name={folder.name!r} path={folder.path}",
                    )
                    continue
                if not self.config.folder_pattern.allowed(folder.name):
                    self.report.folders_filtered += 1
                    self.report.report_filtered(
                        "pattern",
                        "Folder",
                        f"name={folder.name!r} path={folder.path}",
                    )
                    continue
                parent_project = self._resolve_parent_project(folder)
                if parent_project and parent_project not in self._emitted_project_names:
                    # Parent was filtered out upstream; skipping keeps the UI
                    # tree clean instead of surfacing orphan folders.
                    self.report.folders_filtered += 1
                    self.report.report_filtered(
                        "parent-project-filtered",
                        "Folder",
                        f"name={folder.name!r} path={folder.path}",
                    )
                    continue
                self.report.folders_scanned += 1
                parent_project = parent_project or ORPHAN_PROJECT_SENTINEL
                self._emitted_folder_keys.add((parent_project, folder.name))
                logger.debug(
                    "Emitting Folder container: name=%s path=%s parent_project=%s",
                    folder.name,
                    folder.path,
                    parent_project,
                )
                yield self._make_folder_container(folder)
            except (
                InformaticaApiError,
                InformaticaLoginError,
                requests.RequestException,
                ValueError,
                KeyError,
            ) as e:
                self.report.warning(
                    title="Failed to process IDMC folder",
                    message="This folder will be skipped.",
                    context=f"id={folder.id}, name={folder.name!r}",
                    exc=e,
                )

    def _safe_list(
        self,
        iterable_factory: Callable[[], Iterable[T]],
        *,
        title: str,
        message: str,
        context: str,
    ) -> Iterator[T]:
        """Wrap an IDMC listing call, converting enumeration failures to
        report warnings. Takes a zero-arg factory so call-time failures
        (non-generator clients that HTTP-fail on invocation) are caught
        the same as iteration-time failures.
        """
        try:
            yield from iterable_factory()
        except (
            InformaticaApiError,
            InformaticaLoginError,
            requests.RequestException,
            ValueError,
            KeyError,
        ) as e:
            self.report.warning(title=title, message=message, context=context, exc=e)

    def _make_project_container(self, project: IdmcObject) -> Container:
        return Container(
            container_key=self._project_key(project.name),
            display_name=project.name,
            description=project.description,
            subtype="Project",
            owners=self._owner_list(project.created_by, project.updated_by),
            tags=self._tag_list(project.tags),
        )

    def _make_folder_container(self, folder: IdmcObject) -> Container:
        parent_project_name = self._resolve_parent_project(folder)
        # Only reference the project when it was actually emitted; otherwise
        # the folder would dangle with parent_container → a ghost URN.
        parent_key: Optional[ContainerKey] = (
            self._project_key(parent_project_name)
            if parent_project_name
            and parent_project_name in self._emitted_project_names
            else None
        )
        key = self._folder_key(
            parent_project_name or ORPHAN_PROJECT_SENTINEL, folder.name
        )
        return Container(
            container_key=key,
            display_name=folder.name,
            description=folder.description,
            subtype="Folder",
            owners=self._owner_list(folder.created_by, folder.updated_by),
            parent_container=parent_key,
            tags=self._tag_list(folder.tags),
        )

    def _project_key(self, project_name: str) -> InformaticaProjectKey:
        return InformaticaProjectKey(
            platform=PLATFORM,
            instance=self.config.platform_instance,
            env=self.config.env,
            project_name=project_name,
        )

    def _resolve_parent_project(self, obj: IdmcObject) -> str:
        return self._resolve_parent_project_from_path(obj.path)

    def _excluded_by_folder_filter(self, path: str) -> bool:
        """True when the entity's containing folder is rejected by folder_pattern.

        Cascades the folder filter to entities inside a filtered folder —
        users expect ``folder_pattern`` to also exclude mappings/tasks living
        inside folders that don't match. Entities at the project root
        (2-segment path) aren't subject to folder filtering.
        """
        parts = self._parse_idmc_path(path)
        if len(parts) < 3:
            return False
        folder_name = parts[1]
        return not self.config.folder_pattern.allowed(folder_name)

    @staticmethod
    def _parse_idmc_path(path: str) -> List[str]:
        """Split an IDMC path into its meaningful segments.

        Drops empty elements and the leading ``Explore`` token so callers
        work with ``[Project, Folder, Asset]`` uniformly whether the raw
        path arrived as ``/Explore/P/F/A`` or ``P/F/A``.
        """
        return [p for p in path.split("/") if p and p != "Explore"]

    @staticmethod
    def _resolve_parent_project_from_path(path: str) -> str:
        """Extract the project name from an IDMC path like '/Explore/{ProjectName}/...'."""
        parts = InformaticaSource._parse_idmc_path(path)
        return parts[0] if parts else ""

    def _browse_path_entries(self, path: str) -> List[BrowsePathEntryClass]:
        """Return BrowsePathsV2 ancestor entries for an IDMC path.

        Emits the Project (and Folder, when present) as container URN entries so
        the entity nests under its IDMC project/folder containers in the UI.
        For '/Explore/Sales/ETL/my_entity' → [ProjectURN(Sales), FolderURN(ETL)].

        When a path segment corresponds to a container we did NOT emit (filtered
        out, or not a real IDMC entity), a plain-name entry is used instead so
        the UI renders the folder's human-readable name rather than a ghost URN.
        """
        parts = self._parse_idmc_path(path)
        ancestors = parts[:-1] if len(parts) > 1 else []
        if not ancestors:
            return []
        project_name = ancestors[0]
        entries: List[BrowsePathEntryClass] = [
            self._container_or_name_entry(
                project_name in self._emitted_project_names,
                self._project_key(project_name).as_urn(),
                project_name,
            )
        ]
        if len(ancestors) >= 2:
            folder_name = ancestors[1]
            entries.append(
                self._container_or_name_entry(
                    (project_name, folder_name) in self._emitted_folder_keys,
                    self._folder_key(project_name, folder_name).as_urn(),
                    folder_name,
                )
            )
            for extra in ancestors[2:]:
                entries.append(BrowsePathEntryClass(id=extra))
        return entries

    @staticmethod
    def _container_or_name_entry(
        emitted: bool, urn: str, name: str
    ) -> BrowsePathEntryClass:
        """Container URN entry when emitted; plain-name entry otherwise.

        Plain-name entries avoid ghost URNs in the UI navigate tree when the
        container for a path segment isn't part of this ingestion run.
        """
        if emitted:
            return BrowsePathEntryClass(id=urn, urn=urn)
        return BrowsePathEntryClass(id=name)

    def _folder_key(self, project_name: str, folder_name: str) -> InformaticaFolderKey:
        return InformaticaFolderKey(
            platform=PLATFORM,
            instance=self.config.platform_instance,
            env=self.config.env,
            project_name=project_name,
            folder_name=folder_name,
        )

    def _container_parent_key(self, path: str) -> Optional[ContainerKey]:
        """Return the deepest emitted-container key for an IDMC path.

        Used with ``flow._set_container`` so DataFlows carry both
        ``ContainerClass.container`` and ``BrowsePathsV2``. DataHub's
        ``browseV2`` needs the ``container`` pointer to group entities at
        the project root consistently with entities inside a folder.
        Returns None when no ancestor container was emitted.
        """
        parts = self._parse_idmc_path(path)
        if len(parts) < 2:
            return None
        project_name = parts[0]
        if project_name not in self._emitted_project_names:
            return None
        if len(parts) >= 3 and (project_name, parts[1]) in self._emitted_folder_keys:
            return self._folder_key(project_name, parts[1])
        return self._project_key(project_name)

    def _extract_taskflows(self) -> Iterable[Entity]:
        # Two-pass: collect filtered Taskflows, batch-prefetch their step
        # definitions in one v3 Export job, then emit using the warmed cache.
        accepted: List[IdmcObject] = []
        taskflows = self._safe_list(
            lambda: self._iter_with_tags("TASKFLOW"),
            title="Failed to list IDMC taskflows",
            message="Taskflow DataFlow entities will be missing.",
            context="/public/core/v3/objects?type=TASKFLOW",
        )
        for tf in taskflows:
            if self._excluded_by_folder_filter(tf.path):
                self.report.taskflows_filtered += 1
                self.report.report_filtered(
                    "folder-excluded",
                    "TASKFLOW",
                    f"name={tf.name!r} path={tf.path}",
                )
                continue
            if not self.config.taskflow_pattern.allowed(
                f"{tf.path}/{tf.name}" if tf.path else tf.name
            ):
                self.report.taskflows_filtered += 1
                self.report.report_filtered(
                    "pattern",
                    "TASKFLOW",
                    f"name={tf.name!r} path={tf.path}",
                )
                continue
            accepted.append(tf)

        if accepted:
            guids = [tf.id for tf in accepted if tf.id]
            self.client.prefetch_taskflow_definitions(guids)

        for tf in accepted:
            try:
                self.report.taskflows_scanned += 1
                logger.debug(
                    "Emitting Taskflow: name=%s path=%s id=%s",
                    tf.name,
                    tf.path,
                    tf.id,
                )
                flow = self._make_taskflow(tf)
                yield flow
                yield from self._emit_taskflow_steps(tf, flow)
            except (
                InformaticaApiError,
                InformaticaLoginError,
                requests.RequestException,
                ValueError,
                KeyError,
            ) as e:
                self.report.warning(
                    title="Failed to process IDMC taskflow",
                    message="This taskflow will be skipped.",
                    context=f"id={tf.id}, name={tf.name!r}",
                    exc=e,
                )

    def _make_taskflow(self, tf: IdmcObject) -> DataFlow:
        display = tf.name or tf.id
        flow = DataFlow(
            platform=PLATFORM,
            name=_safe_flow_id(tf.name, fallback=tf.id or display),
            platform_instance=self.config.platform_instance,
            env=self.config.env,
            display_name=display,
            description=tf.description,
            custom_properties=self._taskflow_custom_props(tf),
            owners=self._owner_list(tf.created_by, tf.updated_by),
            subtype="Taskflow",
            tags=self._tag_list(tf.tags),
        )
        self._attach_container_and_browse_path(flow, tf.path)
        return flow

    def _attach_container_and_browse_path(
        self, flow: DataFlow, path: str
    ) -> List[BrowsePathEntryClass]:
        """Wire a DataFlow to its parent Container + BrowsePathsV2.

        Returns the entries so callers that build deeper paths (e.g. MT
        DataJob appends the MT name) can extend them.
        """
        parent_key = self._container_parent_key(path)
        if parent_key is not None:
            flow._set_container(parent_key)
        browse_entries = self._browse_path_entries(path) if path else []
        if browse_entries:
            flow._set_aspect(BrowsePathsV2Class(path=browse_entries))
        return browse_entries

    @staticmethod
    def _audit_props(obj: IdmcObject, **extra: str) -> Dict[str, str]:
        """Common ``customProperties`` (v3 id, type, path, audit fields)
        shared across Taskflow / MT emission. Empty strings are stripped.
        """
        props: Dict[str, str] = {
            "v3Id": obj.id,
            "objectType": obj.object_type or "",
            "path": obj.path,
            "createdBy": obj.created_by or "",
            "updatedBy": obj.updated_by or "",
            "createTime": obj.create_time.isoformat() if obj.create_time else "",
            "updateTime": obj.update_time.isoformat() if obj.update_time else "",
        }
        props.update(extra)
        return {k: v for k, v in props.items() if v}

    @staticmethod
    def _taskflow_custom_props(tf: IdmcObject) -> Dict[str, str]:
        return InformaticaSource._audit_props(
            tf, objectType=tf.object_type or "TASKFLOW"
        )

    @staticmethod
    def _mt_custom_props(
        mt: IdmcMappingTask, mapping_name: str, mapping_v3_guid: str
    ) -> Dict[str, str]:
        # Historical quirk: the ``v3Id`` key carries the v2 id (MTT v3
        # GUIDs contain separators the UI would split). Kept for back-compat
        # with previously-emitted entities.
        return {
            k: v
            for k, v in {
                "v3Id": mt.v2_id,
                "objectType": "MTT",
                "path": mt.path,
                "mappingId": mt.mapping_id or "",
                "mappingName": mapping_name,
                "mappingV3Id": mapping_v3_guid,
                "createdBy": mt.created_by or "",
                "updatedBy": mt.updated_by or "",
                "createTime": mt.create_time.isoformat() if mt.create_time else "",
                "updateTime": mt.update_time.isoformat() if mt.update_time else "",
            }.items()
            if v
        }

    def _emit_taskflow_steps(self, tf: IdmcObject, flow: DataFlow) -> Iterable[Entity]:
        """Emit a chained-MT orchestration graph for a Taskflow.

        The resulting Taskflow lineage reads as
        ``input_ds → MT1 → … → MTn → orchestrate → out_ds``. Step-order
        edges between MTs go into ``_mt_predecessors`` (merged into each
        MT's ``inputDatajobs`` during the lineage phase); a single
        ``orchestrate`` DataJob is emitted per Taskflow at the end.

        Non-data steps (command / decision / …) are summarized in
        ``customProperties`` but don't participate in the chain.
        Silently no-ops when the definition isn't available.
        """
        definition = self.client.get_taskflow_definition(
            taskflow_name=tf.name, taskflow_v3_guid=tf.id
        )
        if definition is None or not definition.steps:
            return
        walk = self._walk_taskflow_steps(definition.steps)
        if not walk.mt_urns_in_order:
            return
        self.report.taskflows_with_steps += 1
        yield self._emit_orchestrate_datajob(tf, flow, walk)

    def _walk_taskflow_steps(self, steps: List[TaskflowStep]) -> TaskflowStepWalkResult:
        """Resolve step task refs and register predecessor edges in
        ``_mt_predecessors``. Pure — no DataJob emission.
        """
        step_order: List[str] = []
        step_summary_parts: List[str] = []
        mt_urns_in_order: List[str] = []
        prev_mt_urn: Optional[str] = None
        for step in steps:
            step_order.append(step.step_id)
            mt_urn = self._resolve_task_ref(step)
            if mt_urn is None:
                if step.step_type != "data":
                    step_summary_parts.append(f"{step.step_name} ({step.step_type})")
                continue
            if mt_urn not in mt_urns_in_order:
                mt_urns_in_order.append(mt_urn)
            step_summary_parts.append(
                f"{step.step_name} → {step.task_ref_name or step.task_ref_id}"
            )
            # Register prev_mt → mt_urn. Dedup so an MT appearing in two
            # Taskflows with identical ordering doesn't get duplicate edges.
            if prev_mt_urn and prev_mt_urn != mt_urn:
                predecessors = self._mt_predecessors.setdefault(mt_urn, [])
                if prev_mt_urn not in predecessors:
                    predecessors.append(prev_mt_urn)
            prev_mt_urn = mt_urn
        return TaskflowStepWalkResult(
            step_order=step_order,
            mt_urns_in_order=mt_urns_in_order,
            step_summary_parts=step_summary_parts,
        )

    def _emit_orchestrate_datajob(
        self, tf: IdmcObject, flow: DataFlow, walk: TaskflowStepWalkResult
    ) -> DataJob:
        """Emit the orchestrate DataJob and register its URN for the
        lineage phase to bubble last-MT outputs onto.
        """
        last_mt_urn = walk.mt_urns_in_order[-1]
        custom_props: Dict[str, str] = {
            "mappingTaskCount": str(len(walk.mt_urns_in_order)),
            "stepOrder": " → ".join(walk.step_order),
        }
        if walk.step_summary_parts:
            custom_props["stepSummary"] = " | ".join(walk.step_summary_parts)
        job = DataJob(
            name="orchestrate",
            flow=flow,
            display_name=f"{tf.name}.orchestrate",
            subtype="Taskflow Orchestration",
            description=(
                f"Runs {len(walk.mt_urns_in_order)} Mapping Task(s) in order: "
                f"{' → '.join(walk.step_summary_parts)}"
                if walk.step_summary_parts
                else None
            ),
            custom_properties=custom_props,
        )
        # inputDatajobs = [last MT]; outputDatasets is filled during the
        # lineage phase so Taskflow Lineage ends with ``… → orch → out_ds``.
        job._set_aspect(
            DataJobInputOutputClass(
                inputDatasets=[],
                outputDatasets=[],
                inputDatajobs=[last_mt_urn],
            )
        )
        orchestrate_urn = str(job.urn)
        self._orchestrate_state[orchestrate_urn] = OrchestrateState(
            last_mt_urn=last_mt_urn,
            mt_urns_in_order=list(walk.mt_urns_in_order),
        )
        self._orchestrate_by_last_mt.setdefault(last_mt_urn, []).append(orchestrate_urn)
        return job

    def _resolve_task_ref(self, step: TaskflowStep) -> Optional[str]:
        """Resolve a step task reference to its MT's DataJob URN.

        Prefers v2 id over name. ``None`` if the step references an MT
        that wasn't emitted (e.g. filtered by pattern).
        """
        if step.task_ref_id:
            return self._mt_v2_id_to_job_urn.get(V2Id(step.task_ref_id))
        return None

    def _extract_mappings_and_tasks(self) -> Iterable[Entity]:
        """MT-centric emission.

        Only Mapping Tasks are emitted as first-class entities; the v2
        mapping cache is still loaded so ``mt.mapping_id`` → v3 GUID
        translation works for the lineage export batch. Mappings without
        an MT and Mapplets are never emitted.
        """
        # Cache v2 mappings by both v3 GUID and v2 id — lineage export
        # keys by GUID, the /mttask enrichment keys by v2 id.
        try:
            for m in self.client.list_mappings():
                if m.asset_frs_guid:
                    self._v2_mappings_by_guid[m.asset_frs_guid] = m
                if m.v2_id:
                    self._v2_mappings_by_v2_id[m.v2_id] = m
        except (
            InformaticaApiError,
            InformaticaLoginError,
            requests.RequestException,
        ) as e:
            self.report.warning(
                title="Failed to list IDMC v2 mappings",
                message=(
                    "Mapping Tasks may be missing mapping references and "
                    "lineage may not resolve to the right mapping."
                ),
                context="/api/v2/mapping",
                exc=e,
            )

        mapping_tasks = self._safe_list(
            self.client.list_mapping_tasks,
            title="Failed to list IDMC mapping tasks",
            message="Mapping Tasks will be missing from ingestion output.",
            context="/public/core/v3/objects?type=MTT",
        )
        try:
            for mt in mapping_tasks:
                if self._excluded_by_folder_filter(mt.path):
                    self.report.report_filtered(
                        "folder-excluded",
                        "MTT",
                        f"name={mt.name!r} path={mt.path}",
                    )
                    continue
                if not self.config.mapping_task_pattern.allowed(
                    f"{mt.path}/{mt.name}" if mt.path else mt.name
                ):
                    self.report.mapping_tasks_filtered += 1
                    self.report.report_filtered(
                        "pattern",
                        "MTT",
                        f"name={mt.name!r} path={mt.path}",
                    )
                    continue
                self.report.mapping_tasks_scanned += 1
                logger.debug(
                    "Emitting MappingTask: name=%s path=%s mapping_id=%s",
                    mt.name,
                    mt.path or "<empty>",
                    mt.mapping_id or "<empty>",
                )
                if not mt.path:
                    self.report.report_filtered(
                        "mtt-empty-path",
                        "MTT",
                        f"name={mt.name!r} mapping={mt.mapping_id!r}",
                    )
                yield from self._emit_mapping_task(mt)
        finally:
            self._v2_mappings_by_guid.clear()
            self._v2_mappings_by_v2_id.clear()

    def _emit_mapping_task(self, mt: IdmcMappingTask) -> Iterable[Entity]:
        """Emit a Mapping Task as DataFlow + inner ``transform`` DataJob.

        The inner DataJob is the lineage attachment point — each MT
        referencing a given Mapping receives the same tables (fan-out).
        """
        # v2 mapping_id → v3 GUID so the lineage phase can match export
        # output back to this MT's DataJob.
        v2_mapping = (
            self._v2_mappings_by_v2_id.get(mt.mapping_id) if mt.mapping_id else None
        )
        mapping_v3_guid: V3Guid = (
            v2_mapping.asset_frs_guid if v2_mapping else V3Guid("")
        )
        mapping_name = mt.mapping_name or (v2_mapping.name if v2_mapping else "")
        custom_props = self._mt_custom_props(mt, mapping_name, mapping_v3_guid)
        task_tags = self._tag_list(mt.tags)
        if not mt.v2_id:
            logger.warning(
                "Mapping Task %r has no v2_id; falling back to name for URN, "
                "which may collide with same-named MTs in other folders.",
                mt.name,
            )
        flow = DataFlow(
            platform=PLATFORM,
            name=_safe_flow_id(
                mt.v2_id,
                fallback=f"{mt.path}/{mt.name}" if mt.path else mt.name,
            ),
            platform_instance=self.config.platform_instance,
            env=self.config.env,
            display_name=mt.name,
            description=mt.description,
            custom_properties=custom_props,
            owners=self._owner_list(mt.created_by, mt.updated_by),
            subtype="Mapping Task",
            tags=task_tags,
        )
        browse_entries = self._attach_container_and_browse_path(flow, mt.path)
        yield flow
        yield self._make_mt_transform_datajob(
            mt, flow, custom_props, task_tags, browse_entries, mapping_v3_guid
        )

    def _make_mt_transform_datajob(
        self,
        mt: IdmcMappingTask,
        flow: DataFlow,
        custom_props: Dict[str, str],
        task_tags: Optional[List[TagUrn]],
        browse_entries: List[BrowsePathEntryClass],
        mapping_v3_guid: V3Guid,
    ) -> DataJob:
        """Build the MT's inner ``transform`` DataJob and register its
        URN in the indexes the lineage + Taskflow-step phases consume.
        """
        job = DataJob(
            name=MAPPING_JOB_ID,
            flow=flow,
            description=mt.description,
            display_name=f"{mt.name}.transform",
            custom_properties=custom_props,
            subtype="Task Logic",
            owners=self._owner_list(mt.created_by, mt.updated_by),
            tags=task_tags,
        )
        job._set_aspect(
            BrowsePathsV2Class(
                path=[
                    *browse_entries,
                    BrowsePathEntryClass(id=mt.name, urn=str(flow.urn)),
                ]
            )
        )
        job_urn = str(job.urn)
        if mapping_v3_guid:
            self._mapping_v3_to_mt_job_urns.setdefault(mapping_v3_guid, []).append(
                job_urn
            )
            if mapping_v3_guid not in self._mapping_ids:
                self._mapping_ids.append(mapping_v3_guid)
        if mt.v2_id:
            self._mt_v2_id_to_job_urn[mt.v2_id] = job_urn
        return job

    def _extract_lineage(self) -> Iterable[Union[MetadataWorkUnit, Entity]]:
        self._load_connections()
        if not self._connections_by_id:
            # ``_load_connections`` already reported the failure; don't
            # waste time polling export jobs we can't resolve to URNs.
            logger.warning("No connections loaded; skipping lineage extraction.")
            return
        if not self._mapping_ids:
            logger.info("No mappings to extract lineage for.")
            return
        batch_size = self.config.export_batch_size
        batches = [
            (self._mapping_ids[i : i + batch_size], i)
            for i in range(0, len(self._mapping_ids), batch_size)
        ]
        # submit+wait runs concurrently; download+parse+emit stays on main thread.
        future_to_batch: Dict[Future[str], Tuple[List[V3Guid], int]] = {}
        with ThreadPoolExecutor(
            max_workers=self.config.max_concurrent_export_jobs
        ) as executor:
            for batch, batch_start in batches:
                future_to_batch[
                    executor.submit(self._submit_and_wait_export, batch, batch_start)
                ] = (batch, batch_start)
            for future in as_completed(future_to_batch):
                batch, batch_start = future_to_batch[future]
                try:
                    job_id = future.result()
                except (
                    InformaticaApiError,
                    InformaticaLoginError,
                    requests.RequestException,
                ) as e:
                    self.report.warning(
                        title="IDMC export batch failed",
                        message="Lineage for this batch will be missing.",
                        context=f"batch_start={batch_start}, size={len(batch)}",
                        exc=e,
                    )
                    self.report.export_jobs_failed.append(f"batch@{batch_start}: {e}")
                    continue
                if job_id:
                    for lineage_info in self.client.download_and_parse_export(
                        job_id, batch
                    ):
                        yield from self._emit_lineage(lineage_info)
        # Final pass: bubble each last-MT's outputs up to its Taskflow's
        # orchestrate so the Taskflow Lineage view has the full pipeline.
        yield from self._emit_orchestrate_aggregated_lineage()

    def _submit_and_wait_export(self, batch: List[V3Guid], batch_start: int) -> str:
        """Submit one export batch and block until complete. Returns job_id on
        success, empty string on non-success (warning already recorded)."""
        job_id = self.client.submit_export_job(batch)
        status = self.client.wait_for_export(job_id)
        if status.state != ExportJobState.SUCCESSFUL:
            if status.state != ExportJobState.TIMEOUT:
                self.report.warning(
                    title="IDMC export job did not succeed",
                    message="Lineage for this batch will be missing.",
                    context=(
                        f"job_id={job_id}, state={status.state.value}, "
                        f"batch_start={batch_start}, "
                        f"message={status.message or '<none>'}"
                    ),
                )
            return ""
        return job_id

    def _emit_orchestrate_aggregated_lineage(self) -> Iterable[MetadataWorkUnit]:
        """Emit each orchestrate's final ``DataJobInputOutput`` with the
        last MT's outputs mirrored.
        """
        for orch_urn, state in self._orchestrate_state.items():
            outputs = state.output_datasets
            if not outputs:
                # No resolved outputs on the last MT — keep the base
                # emission with inputDatajobs=[last_mt] and no edges.
                continue
            yield MetadataChangeProposalWrapper(
                entityUrn=orch_urn,
                aspect=DataJobInputOutputClass(
                    inputDatasets=[],
                    outputDatasets=sorted(outputs),
                    inputDatajobs=[state.last_mt_urn],
                ),
            ).as_workunit()

    def _load_connections(self) -> None:
        """Load all connections into cache, keyed by federatedId and id (separate dicts)."""
        if self._connections_by_id or self._connections_by_fed_id:
            return
        try:
            for conn in self.client.list_connections():
                if conn.federated_id:
                    self._connections_by_fed_id[conn.federated_id] = conn
                self._connections_by_id[conn.id] = conn
        except (
            InformaticaApiError,
            InformaticaLoginError,
            requests.RequestException,
        ) as e:
            self.report.failure(
                title="Failed to load IDMC connections",
                message="Lineage resolution will be degraded; upstream/downstream datasets cannot be resolved.",
                context="/api/v2/connection",
                exc=e,
            )

    def _resolve_connection_platform(self, connection: IdmcConnection) -> Optional[str]:
        """Resolve an IDMC connection to a DataHub platform name.

        Priority (first hit wins):
          1. ``connection_type_overrides[id]`` — per-connection user override
          2. ``connection_type_platform_map[conn_type]`` — recipe-level map
          3. Built-in ``CONNECTION_TYPE_MAP[conn_type]``
          4. ``connection_type_platform_map[base_type]``
          5. Built-in ``CONNECTION_TYPE_MAP[base_type]``
          6. Substring inference from the connection's display name (for
             IDMC orgs that return ``conn_type=''`` ``base_type='TOOLKIT'``
             on marketplace connectors).

        ``conn_type`` is ``connParams["Connection Type"]`` (specific);
        ``base_type`` is the generic ``type`` field (often ``TOOLKIT``).
        """
        override = self.config.connection_type_overrides.get(connection.id)
        if override:
            return override
        user_map = self.config.connection_type_platform_map
        # First-hit-wins lookup chain; user map wins over built-in, and
        # conn_type (specific) wins over base_type (generic ``TOOLKIT``).
        lookups = (
            (connection.conn_type, user_map),
            (connection.conn_type, CONNECTION_TYPE_MAP),
            (connection.base_type, user_map),
            (connection.base_type, CONNECTION_TYPE_MAP),
        )
        for key, mapping in lookups:
            if key and key in mapping:
                return mapping[key]
        inferred = self._platform_from_connection_name(connection.name)
        if inferred:
            logger.info(
                "Inferred platform=%s for connection id=%s name=%r via "
                "name-substring fallback (conn_type=%r base_type=%r)",
                inferred,
                connection.id,
                connection.name,
                connection.conn_type,
                connection.base_type,
            )
            return inferred
        return None

    @staticmethod
    def _platform_from_connection_name(name: str) -> Optional[str]:
        """Best-effort platform inference from a connection's display name.

        Scans CONNECTION_TYPE_MAP keys for any that appear as a
        case-insensitive substring of the name (with non-alphanumerics
        stripped). Returns the platform for the *longest* matching key so
        more-specific identifiers win over shorter ones.
        """
        if not name:
            return None
        haystack = "".join(c for c in name.lower() if c.isalnum())
        best_needle_len = 0
        best_platform: Optional[str] = None
        for key, platform in CONNECTION_TYPE_MAP.items():
            needle = "".join(c for c in key.lower() if c.isalnum())
            if needle and needle in haystack:
                if len(needle) > best_needle_len:
                    best_needle_len = len(needle)
                    best_platform = platform
        return best_platform

    def _emit_lineage(
        self, lineage_info: MappingLineageInfo
    ) -> Iterable[MetadataWorkUnit]:
        """Fan out ``dataJobInputOutput`` + dataset-level ``upstreamLineage``.

        One Mapping can be referenced by N Mapping Tasks, so a single export
        result is applied to every MT's ``transform`` DataJob that runs it.
        Dataset-level ``upstreamLineage`` on downstream tables is emitted once
        (the upstream set is the same regardless of which MT triggered it).
        """
        if not lineage_info.mapping_id:
            self.report.warning(
                title="IDMC lineage missing mapping id",
                message="Could not align mapping export entry to a submitted v3 GUID; skipping lineage emission.",
                context=f"mapping_name={lineage_info.mapping_name}",
            )
            return
        mt_job_urns = self._mapping_v3_to_mt_job_urns.get(lineage_info.mapping_id)
        if not mt_job_urns:
            self.report.warning(
                title="IDMC lineage references a mapping with no Mapping Tasks",
                message="No Mapping Task was emitted that runs this mapping; lineage will be skipped.",
                context=f"mapping_id={lineage_info.mapping_id} name={lineage_info.mapping_name}",
            )
            return
        input_urns: List[str] = [
            urn
            for src in lineage_info.source_tables
            if (urn := self._resolve_table_to_dataset_urn(src)) is not None
        ]
        output_urns: List[str] = [
            urn
            for tgt in lineage_info.target_tables
            if (urn := self._resolve_table_to_dataset_urn(tgt)) is not None
        ]
        if not input_urns and not output_urns:
            return
        yield from self._stub_external_datasets(input_urns, output_urns)
        for job_urn in mt_job_urns:
            yield from self._emit_mt_datajob_io(job_urn, input_urns, output_urns)
        yield from self._emit_downstream_upstream_lineage(input_urns, output_urns)

    def _stub_external_datasets(
        self, input_urns: List[str], output_urns: List[str]
    ) -> Iterable[MetadataWorkUnit]:
        """Emit a minimal ``Status`` aspect per unseen external dataset URN.

        Without this stub, DataHub treats URNs that no other connector has
        ingested as non-existent and ``searchAcrossLineage`` filters them
        out of results. ``UpstreamLineage`` on a downstream creates a key
        for *output* URNs implicitly, but inputs are only referenced
        inside aspects and would otherwise remain ``exists: false``.

        Dedup'd per-source-run via ``_stubbed_external_dataset_urns``.
        """
        for ds_urn in (*input_urns, *output_urns):
            if ds_urn in self._stubbed_external_dataset_urns:
                continue
            self._stubbed_external_dataset_urns.add(ds_urn)
            yield MetadataChangeProposalWrapper(
                entityUrn=ds_urn,
                aspect=StatusClass(removed=False),
            ).as_workunit()

    def _emit_mt_datajob_io(
        self,
        job_urn: str,
        input_urns: List[str],
        output_urns: List[str],
    ) -> Iterable[MetadataWorkUnit]:
        """Emit one MT ``transform`` DataJob's ``DataJobInputOutput`` aspect
        and, when this MT is the last step of any Taskflow, bubble its
        outputs up to the orchestrate's accumulator.
        """
        predecessor_mt_urns = self._mt_predecessors.get(job_urn, [])
        yield MetadataChangeProposalWrapper(
            entityUrn=job_urn,
            aspect=DataJobInputOutputClass(
                inputDatasets=input_urns,
                outputDatasets=output_urns,
                inputDatajobs=predecessor_mt_urns,
            ),
        ).as_workunit()
        self.report.lineage_edges_emitted += len(input_urns) + len(output_urns)
        # When this MT is the last step of any Taskflow, push its outputs
        # onto the orchestrate's accumulated ``outputDatasets``.
        for orch_urn in self._orchestrate_by_last_mt.get(job_urn, []):
            state = self._orchestrate_state.get(orch_urn)
            if state is not None:
                state.output_datasets.update(output_urns)

    def _emit_downstream_upstream_lineage(
        self, input_urns: List[str], output_urns: List[str]
    ) -> Iterable[MetadataWorkUnit]:
        """Emit ``UpstreamLineage`` on each target dataset.

        Emitted alongside the MT's ``DataJobInputOutput`` (not instead of)
        because DataHub's UI lineage traversal has asymmetric edge-type
        filters: a Dataset→Dataset Lineage tab doesn't reliably traverse
        ``Consumes``/``Produces`` edges through an intermediate DataJob.
        Keeping the Dataset-to-Dataset ``UpstreamLineage`` aspect covers
        that path explicitly; the DataJob aspect covers the MT's own
        Lineage tab. The two together render correctly in every view we
        tested.
        """
        if not input_urns:
            return
        for tgt_urn in output_urns:
            yield MetadataChangeProposalWrapper(
                entityUrn=tgt_urn,
                aspect=UpstreamLineageClass(
                    upstreams=[
                        UpstreamClass(
                            dataset=src_urn,
                            type=DatasetLineageTypeClass.TRANSFORMED,
                        )
                        for src_urn in input_urns
                    ]
                ),
            ).as_workunit()

    def _resolve_table_to_dataset_urn(self, table: LineageTable) -> Optional[str]:
        if not table.table_name:
            # Malformed IDMC transformations can yield empty table names; emitting
            # a URN like ``snowflake,DWH.SALES.,PROD`` creates a dangling edge.
            self.report.report_connection_unresolved(
                table.connection_federated_id,
                "unknown",
                "Empty table name in IDMC transformation",
            )
            return None
        conn = self._connections_by_fed_id.get(table.connection_federated_id)
        if not conn:
            self.report.report_connection_unresolved(
                table.connection_federated_id,
                "unknown",
                f"Connection not found for table {table.table_name}",
            )
            return None
        platform = self._resolve_connection_platform(conn)
        if not platform:
            self.report.report_connection_unresolved(
                conn.id,
                conn.name,
                (
                    f"Unmapped connection: conn_type={conn.conn_type!r} "
                    f"base_type={conn.base_type!r}. Add "
                    f"connection_type_platform_map: {{{conn.conn_type or conn.base_type!r}: "
                    f"'<platform>'}} to the recipe."
                ),
            )
            return None
        self.report.connections_resolved += 1
        parts: List[str] = []
        if conn.database:
            parts.append(conn.database)
        if table.schema_name:
            parts.append(table.schema_name)
        elif conn.db_schema:
            parts.append(conn.db_schema)
        parts.append(table.table_name)
        qualifier = ".".join(parts)
        if self.config.convert_urns_to_lowercase:
            qualifier = qualifier.lower()
        platform_instance = self.config.connection_to_platform_instance.get(conn.id)
        return str(
            DatasetUrn.create_from_ids(
                platform_id=platform,
                table_name=qualifier,
                env=self.config.env,
                platform_instance=platform_instance,
            )
        )

    def _iter_with_tags(self, object_type: IdmcObjectType) -> Iterable[IdmcObject]:
        # Single pass per type; tag matching is client-side. Passing each tag
        # separately would cause K × N API calls since IDMC ignores server-side
        # tag filters for non-Project/Folder types.
        tag_set = (
            set(self.config.tag_filter_names)
            if self.config.tag_filter_names and object_type not in ("Project", "Folder")
            else None
        )
        for obj in self.client.list_objects(object_type):
            if self._is_bundle_object(obj):
                self.report.report_filtered("bundle", object_type, obj.path)
                logger.debug(
                    "Skipping bundle %s: id=%s path=%s",
                    object_type,
                    obj.id,
                    obj.path,
                )
                continue
            if tag_set and not any(t in obj.tags for t in tag_set):
                self.report.report_filtered("tag", object_type, obj.path)
                continue
            yield obj

    @staticmethod
    def _is_project_masquerading_as_folder(obj: IdmcObject) -> bool:
        # IDMC's ``type=Folder`` query sometimes returns project roots
        # (path ``/Explore/<name>``) — emitting those would duplicate the Project.
        return len(InformaticaSource._parse_idmc_path(obj.path)) <= 1

    @staticmethod
    def _is_bundle_object(obj: IdmcObject) -> bool:
        return obj.is_bundle()

    def _owner_list(
        self,
        created_by: Optional[str] = None,
        updated_by: Optional[str] = None,
    ) -> Optional[List[Tuple[CorpUserUrn, str]]]:
        # Creator → DATAOWNER, last updater → TECHNICAL_OWNER; dedup
        # when they're the same principal.
        if not self.config.extract_ownership:
            return None
        created_urn = self._user_urn(created_by)
        updated_urn = self._user_urn(updated_by)
        owners: List[Tuple[CorpUserUrn, str]] = []
        if created_urn:
            owners.append((created_urn, OwnershipTypeClass.DATAOWNER))
        if updated_urn and updated_urn != created_urn:
            owners.append((updated_urn, OwnershipTypeClass.TECHNICAL_OWNER))
        return owners or None

    def _user_urn(self, identifier: Optional[str]) -> Optional[CorpUserUrn]:
        if not identifier:
            return None
        username = identifier
        if self.config.strip_user_email_domain and "@" in username:
            username = username.split("@", 1)[0]
        return CorpUserUrn(username)

    def _tag_list(self, tags: List[str]) -> Optional[List[TagUrn]]:
        if not tags or not self.config.extract_tags:
            return None
        return [TagUrn(name=t) for t in tags]



"""Data pack registry with remote discovery and bundled fallback."""

import importlib.resources
import json
import logging
import os
import pathlib
import time
from typing import Dict, Optional

import requests

from datahub.cli.config_utils import DATAHUB_ROOT_FOLDER
from datahub.cli.datapack.models import DataPackInfo, RegistryManifest

logger = logging.getLogger(__name__)

REGISTRY_URL_ENV = "DATAHUB_DATAPACK_REGISTRY_URL"
DEFAULT_REGISTRY_URL = "https://raw.githubusercontent.com/datahub-project/datahub/master/metadata-ingestion/src/datahub/cli/datapack/resources/registry.json"

REGISTRY_CACHE_PATH = os.path.join(DATAHUB_ROOT_FOLDER, "datapack-registry-cache.json")
REGISTRY_CACHE_TTL_SECONDS = 3600  # 1 hour


def _load_bundled_registry() -> RegistryManifest:
    """Load the registry JSON bundled with the Python package."""
    ref = importlib.resources.files("datahub.cli.datapack.resources").joinpath(
        "registry.json"
    )
    data = json.loads(ref.read_text(encoding="utf-8"))
    return RegistryManifest.model_validate(data)


def _get_registry_url() -> str:
    return os.environ.get(REGISTRY_URL_ENV, DEFAULT_REGISTRY_URL)


def _read_cache() -> Optional[RegistryManifest]:
    """Read cached registry if it exists and is not expired."""
    cache_path = pathlib.Path(REGISTRY_CACHE_PATH)
    if not cache_path.exists():
        return None

    try:
        stat = cache_path.stat()
        age_seconds = time.time() - stat.st_mtime
        if age_seconds > REGISTRY_CACHE_TTL_SECONDS:
            logger.debug("Registry cache expired (age=%.0fs)", age_seconds)
            return None

        with open(cache_path) as f:
            data = json.load(f)
        return RegistryManifest.model_validate(data)
    except Exception:
        logger.debug("Failed to read registry cache", exc_info=True)
        return None


def _write_cache(manifest: RegistryManifest) -> None:
    """Write registry manifest to local cache."""
    cache_path = pathlib.Path(REGISTRY_CACHE_PATH)
    try:
        cache_path.parent.mkdir(parents=True, exist_ok=True)
        with open(cache_path, "w") as f:
            json.dump(manifest.model_dump(), f)
    except Exception:
        logger.debug("Failed to write registry cache", exc_info=True)


def _fetch_remote_registry() -> Optional[RegistryManifest]:
    """Fetch the registry JSON from the remote URL."""
    url = _get_registry_url()
    try:
        logger.debug("Fetching data pack registry from %s", url)
        response = requests.get(url, timeout=15)
        response.raise_for_status()
        data = response.json()
        manifest = RegistryManifest.model_validate(data)
        _write_cache(manifest)
        return manifest
    except Exception:
        logger.debug("Failed to fetch remote registry from %s", url, exc_info=True)
        return None


def fetch_registry(no_cache: bool = False) -> RegistryManifest:
    """Fetch the data pack registry, falling back through layers.

    Resolution order:
    1. Local cache (if fresh and no_cache=False)
    2. Remote registry URL
    3. Bundled registry.json (shipped with the package)

    Args:
        no_cache: If True, skip the local cache and always fetch remote.

    Returns:
        The registry manifest.
    """
    # Try cache first
    if not no_cache:
        cached = _read_cache()
        if cached is not None:
            logger.debug("Using cached registry (%d packs)", len(cached.packs))
            return cached

    # Try remote
    remote = _fetch_remote_registry()
    if remote is not None:
        logger.debug("Using remote registry (%d packs)", len(remote.packs))
        return remote

    # Fall back to bundled
    bundled = _load_bundled_registry()
    logger.debug("Using bundled fallback registry (%d packs)", len(bundled.packs))
    return bundled


def list_packs(no_cache: bool = False) -> Dict[str, DataPackInfo]:
    """List all known data packs."""
    registry = fetch_registry(no_cache=no_cache)
    return dict(registry.packs)


def get_pack(name: str, no_cache: bool = False) -> DataPackInfo:
    """Look up a data pack by name.

    Raises:
        click.UsageError: If the pack is not found.
    """
    import click

    packs = list_packs(no_cache=no_cache)
    if name not in packs:
        available = ", ".join(sorted(packs.keys()))
        raise click.UsageError(
            f"Unknown data pack '{name}'. Available packs: {available}"
        )
    return packs[name]



"""PostgreSQL-specific profiling adapter."""

import logging
from typing import Any, List, Optional

import sqlalchemy as sa
from sqlalchemy.engine import Connection
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.sql.elements import ColumnElement

from datahub.ingestion.source.sqlalchemy_profiler.base_adapter import (
    DEFAULT_QUANTILES,
    PlatformAdapter,
)

logger = logging.getLogger(__name__)


class PostgresAdapter(PlatformAdapter):
    """
    PostgreSQL-specific profiling adapter.

    PostgreSQL features:
    1. Fast row count estimation via pg_class.reltuples
    2. Standard SQL for most operations
    3. No native MEDIAN function (would need custom implementation)

    Uses default setup_profiling and cleanup from PlatformAdapter.
    """

    # =========================================================================
    # SQL Expression Builders
    # =========================================================================

    def get_approx_unique_count_expr(self, column: str) -> ColumnElement[Any]:
        """
        PostgreSQL approximate unique count - uses exact COUNT(DISTINCT).

        PostgreSQL doesn't have a built-in approximate count distinct function,
        so we fall back to exact count.

        Args:
            column: Column name

        Returns:
            SQLAlchemy expression for COUNT(DISTINCT column)
        """
        return sa.func.count(sa.func.distinct(sa.column(column)))

    def get_median_expr(self, column: str) -> Optional[ColumnElement[Any]]:
        """
        PostgreSQL median via PERCENTILE_CONT.

        PostgreSQL doesn't have a native MEDIAN function, but supports
        PERCENTILE_CONT which can compute the median (0.5 percentile).

        Args:
            column: Column name (from database schema, already validated)

        Returns:
            SQLAlchemy expression for PERCENTILE_CONT(0.5)
        """
        quoted_column = self.quote_identifier(column)
        # IMPORTANT: label() is required whenever using literal_column()
        # Without label(), the column name would be the entire SQL string, which breaks query combiner result extraction
        return sa.literal_column(
            f"PERCENTILE_CONT(0.5) WITHIN GROUP (ORDER BY {quoted_column})"
        ).label("median")

    def get_column_quantiles(
        self,
        table: sa.Table,
        column: str,
        conn: Connection,
        quantiles: Optional[List[float]] = None,
    ) -> List[Optional[float]]:
        """
        PostgreSQL quantiles via PERCENTILE_DISC.

        Matches GE behavior (sqlalchemy_dataset.py:_get_column_quantiles_generic_sqlalchemy)
        which uses PERCENTILE_DISC(q) WITHIN GROUP (ORDER BY col ASC). The base adapter's
        PERCENTILE_CONT interpolates between values; PERCENTILE_DISC returns the actual
        value at the discrete percentile. These return different values for sparse data.
        """
        if quantiles is None:
            quantiles = DEFAULT_QUANTILES

        for q in quantiles:
            if not (0 <= q <= 1):
                raise ValueError(
                    f"Quantiles must be in [0, 1], got {q}. "
                    f"Quantiles represent percentiles as decimals (e.g., 0.5 for median)."
                )

        quoted_column = self.quote_identifier(column)
        results: List[Optional[float]] = []
        for q in quantiles:
            try:
                percentile_expr = sa.literal_column(
                    f"PERCENTILE_DISC({q}) WITHIN GROUP (ORDER BY {quoted_column} ASC)"
                ).label("percentile")
                query = sa.select([percentile_expr]).select_from(table)
                result = conn.execute(query).scalar()
                results.append(float(result) if result is not None else None)
            except SQLAlchemyError as e:
                logger.warning(
                    f"Failed to compute quantile {q} for {column}: "
                    f"{type(e).__name__}: {str(e)}"
                )
                results.append(None)
        return results

    # =========================================================================
    # Row Count Estimation
    # =========================================================================

    def supports_row_count_estimation(self) -> bool:
        """
        PostgreSQL supports fast row count estimation.

        Returns:
            True - PostgreSQL has pg_class.reltuples
        """
        return True

    def get_estimated_row_count(
        self, table: sa.Table, conn: Connection
    ) -> Optional[int]:
        """
        Get fast row count estimate using pg_class.reltuples.

        This avoids a full table scan by using PostgreSQL's statistics.
        The estimate may be slightly out of date but is very fast.

        Args:
            table: SQLAlchemy table object
            conn: Active database connection

        Returns:
            Estimated row count, or None if query fails
        """
        try:
            schema = table.schema
            table_name = table.name

            # Query pg_class and pg_namespace directly using SQLAlchemy query builder
            pg_class = sa.Table(
                "pg_class",
                sa.MetaData(),
                sa.Column("relname", sa.String),
                sa.Column("relnamespace", sa.Integer),
                sa.Column("reltuples", sa.Float),
                schema="pg_catalog",
            )
            pg_namespace = sa.Table(
                "pg_namespace",
                sa.MetaData(),
                sa.Column("oid", sa.Integer),
                sa.Column("nspname", sa.String),
                schema="pg_catalog",
            )
            query = (
                sa.select([sa.cast(pg_class.c.reltuples, sa.BigInteger)])
                .select_from(
                    pg_class.join(
                        pg_namespace, pg_class.c.relnamespace == pg_namespace.c.oid
                    )
                )
                .where(pg_namespace.c.nspname == schema)
                .where(pg_class.c.relname == table_name)
            )

            result = conn.execute(query).scalar()
            return int(result) if result is not None else None

        except SQLAlchemyError as e:
            logger.debug(
                f"Failed to get PostgreSQL row count estimate: {type(e).__name__}: {str(e)}"
            )
            return None



"""Incremental ownership workunit processor.

Two behaviors diverge intentionally from :func:`auto_incremental_properties`:

1. Dataset-only entity filter. ``OwnershipClass`` is attached to ~28 entity
   types (datasets, containers, charts, dashboards, glossary terms, ML models,
   …). The patch path uses ``create_dataset_owners_patch_builder`` →
   ``DatasetPatchBuilder``, which is dataset-specific. We therefore filter
   explicitly by dataset URN (MCE) / ``entityType == "dataset"`` (MCPW);
   non-dataset ownership passes through as UPSERT. Compare
   ``auto_incremental_lineage``, which dispatches across
   dataset/chart/dashboard patch builders for the same reason.

2. Empty owners are dropped silently. ``create_dataset_owners_patch_builder``
   emits one patch op per owner, so an empty ``owners`` list yields zero ops
   and ``build()`` returns ``[]``. The properties helper does not hit this
   case because ``create_dataset_props_patch_builder`` unconditionally sets
   seven fields, producing at least one op regardless of input. The drop here
   is a no-op for incremental semantics (no ops = no change).
"""

import logging
from typing import Iterable, Optional

from pydantic.fields import Field

from datahub.configuration.common import ConfigModel
from datahub.emitter.mce_builder import set_aspect
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.source_helpers import create_dataset_owners_patch_builder
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.schema_classes import (
    MetadataChangeEventClass,
    OwnershipClass,
    SystemMetadataClass,
)

logger = logging.getLogger(__name__)


def convert_ownership_to_patch(
    urn: str,
    aspect: OwnershipClass,
    system_metadata: Optional[SystemMetadataClass] = None,
) -> MetadataWorkUnit:
    patch_builder = create_dataset_owners_patch_builder(urn, aspect, system_metadata)
    mcp = next(iter(patch_builder.build()))
    return MetadataWorkUnit(id=MetadataWorkUnit.generate_workunit_id(mcp), mcp_raw=mcp)


def auto_incremental_ownership(
    incremental_ownership: bool,
    stream: Iterable[MetadataWorkUnit],
) -> Iterable[MetadataWorkUnit]:
    if not incremental_ownership:
        yield from stream
        return  # early exit

    for wu in stream:
        urn = wu.get_urn()

        if isinstance(wu.metadata, MetadataChangeEventClass):
            if urn.startswith("urn:li:dataset:"):
                ownership_aspect = wu.get_aspect_of_type(OwnershipClass)
                set_aspect(wu.metadata, None, OwnershipClass)
                if len(wu.metadata.proposedSnapshot.aspects) > 0:
                    yield wu

                if ownership_aspect and ownership_aspect.owners:
                    yield convert_ownership_to_patch(
                        urn, ownership_aspect, wu.metadata.systemMetadata
                    )
            else:
                yield wu
        elif (
            isinstance(wu.metadata, MetadataChangeProposalWrapper)
            and isinstance(wu.metadata.aspect, OwnershipClass)
            and wu.metadata.entityType == "dataset"
        ):
            ownership_aspect = wu.metadata.aspect
            if ownership_aspect.owners:
                yield convert_ownership_to_patch(
                    urn, ownership_aspect, wu.metadata.systemMetadata
                )
        else:
            yield wu


class IncrementalOwnershipConfigMixin(ConfigModel):
    incremental_ownership: bool = Field(
        default=False,
        description="When enabled, emits ownership as incremental to existing ownership "
        "in DataHub. When disabled, re-states ownership on each run.",
    )



"""psycopg2 connection factory with optional RDS IAM authentication.

This module intentionally avoids importing ``datahub.ingestion.source.*`` at import time.
RDS IAM support lazy-imports ``AwsConnectionConfig`` / ``RDSIAMTokenManager`` only when
``auth_mode`` is :class:`~datahub.pgqueue.config.PgQueueAuthMode.AWS_IAM`, mirroring the
Java-side isolation of pgQueue's DB access from general metadata ingestion.
"""

from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING, Optional

from datahub.pgqueue.config import PgQueueAuthMode, PgQueueConnectionConfig
from datahub.pgqueue.host_port import parse_host_port

if TYPE_CHECKING:
    from psycopg2.extensions import connection as PGConnection

logger = logging.getLogger(__name__)


def create_pgqueue_connection(config: PgQueueConnectionConfig) -> "PGConnection":
    """Open a new PostgreSQL connection suitable for pgQueue operations."""
    import psycopg2

    kwargs = build_psycopg2_connect_kwargs(config)
    conn = psycopg2.connect(**kwargs)
    return conn


def flush_pg_connection(conn: "PGConnection") -> None:
    """Align connection transaction state at flush boundaries (Kafka sink parity).

    Repository helpers disable autocommit, commit, then restore the prior flag. If an
    error leaves the session *idle in failed transaction*, toggling autocommit can
    fail until rollback — callers should use :func:`restore_pg_connection_autocommit`
    after each block; this function commits an open transaction or rolls back a
    failed one at work-unit boundaries (see :meth:`DatahubPgQueueEmitter.flush`).
    """
    if getattr(conn, "closed", 0):
        return
    try:
        from psycopg2 import extensions
    except ImportError:
        return
    try:
        status = conn.get_transaction_status()
    except Exception:
        return
    if status == extensions.TRANSACTION_STATUS_INERROR:
        try:
            conn.rollback()
        except Exception:
            logger.debug("flush_pg_connection: rollback failed", exc_info=True)
        return
    if status == extensions.TRANSACTION_STATUS_INTRANS:
        try:
            conn.commit()
        except Exception:
            logger.warning("flush_pg_connection: commit failed", exc_info=True)
            try:
                conn.rollback()
            except Exception:
                logger.debug(
                    "flush_pg_connection: rollback after failed commit", exc_info=True
                )


def restore_pg_connection_autocommit(
    conn: "PGConnection", saved_autocommit: bool
) -> None:
    """Restore ``autocommit`` after a transactional block; rollback if the session aborted."""
    try:
        from psycopg2 import extensions

        if conn.get_transaction_status() == extensions.TRANSACTION_STATUS_INERROR:
            conn.rollback()
    except Exception:
        logger.debug(
            "restore_pg_connection_autocommit: pre-restore check failed",
            exc_info=True,
        )
    try:
        conn.autocommit = saved_autocommit
    except Exception:
        logger.debug("restore_pg_connection_autocommit: failed", exc_info=True)


def build_psycopg2_connect_kwargs(config: PgQueueConnectionConfig) -> dict:
    host, port = parse_host_port(config.host_port, default_port=5432)
    if port is None:
        raise ValueError(f"Could not parse port from host_port={config.host_port!r}")

    sslmode = config.sslmode
    password: Optional[str] = None

    if config.auth_mode == PgQueueAuthMode.PASSWORD:
        assert config.password is not None
        password = config.password.get_secret_value()
    elif config.auth_mode == PgQueueAuthMode.AWS_IAM:
        from datahub.ingestion.source.aws.aws_common import (
            AwsConnectionConfig,
            RDSIAMTokenManager,
        )

        aws_cfg = AwsConnectionConfig.model_validate(config.aws_config)
        mgr = RDSIAMTokenManager(
            endpoint=host,
            username=config.username,
            port=port,
            aws_config=aws_cfg,
        )
        password = mgr.get_token()
        if sslmode not in ("require", "verify-ca", "verify-full"):
            sslmode = "require"

    return {
        "host": host,
        "port": port,
        "dbname": config.database,
        "user": config.username,
        "password": password,
        "sslmode": sslmode,
        "connect_timeout": config.connect_timeout_seconds,
    }


def consumer_lock_owner(consumer_group: str) -> str:
    """Unique lock_owner string per consumer process.

    Matches the Java format in PgQueuePollWorker: ``{consumerGroupId}:{UUID}``.
    """
    return f"{consumer_group}:{uuid.uuid4()}"



import logging
from typing import Dict, Optional

from datahub.emitter.mce_builder import (
    datahub_guid,
    make_container_urn,
    make_data_flow_urn,
    make_data_job_urn_with_flow,
    make_data_process_instance_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
)
from datahub.ingestion.source.matillion_dpc.config import MatillionSourceConfig
from datahub.ingestion.source.matillion_dpc.constants import (
    MATILLION_PLATFORM,
)
from datahub.ingestion.source.matillion_dpc.models import (
    MatillionDatasetInfo,
    MatillionEnvironment,
    MatillionPipeline,
    MatillionProject,
)

logger = logging.getLogger(__name__)

# Pipeline file suffixes for name normalization
PIPELINE_FILE_SUFFIXES = [".orch.yaml", ".tran.yaml", ".yaml", ".yml"]


# Standalone utility functions


def extract_base_pipeline_name(job_name: str) -> str:
    """
    Extract base pipeline name from job name with path and extension.

    Examples:
        "folder/pipeline.tran.yaml" -> "pipeline"
        "pipeline.orch.yaml" -> "pipeline"
        "simple-pipeline" -> "simple-pipeline"
    """
    # Extract filename from path
    base_name = job_name.split("/")[-1] if "/" in job_name else job_name

    # Strip pipeline file extensions
    for suffix in PIPELINE_FILE_SUFFIXES:
        if base_name.endswith(suffix):
            return base_name[: -len(suffix)]
    return base_name


def normalize_pipeline_name(pipeline_name: str) -> str:
    """
    Normalize pipeline name by stripping file extensions.

    Examples:
        "pipeline.tran.yaml" -> "pipeline"
        "pipeline" -> "pipeline"
    """
    for suffix in PIPELINE_FILE_SUFFIXES:
        if pipeline_name.endswith(suffix):
            return pipeline_name[: -len(suffix)]
    return pipeline_name


def match_pipeline_name(published_name: str, job_name: str) -> bool:
    """
    Check if a published pipeline name matches a job name from lineage events.

    Handles various formats:
    - Exact match
    - With/without path prefixes
    - With/without file extensions
    """
    if published_name == job_name:
        return True

    base_job_name = extract_base_pipeline_name(job_name)
    if published_name == base_job_name:
        return True

    normalized_job = normalize_pipeline_name(job_name)
    if published_name == normalized_job:
        return True

    return False


def make_step_dpi_urn(
    config: MatillionSourceConfig,
    project_id: str,
    pipeline_name: str,
    execution_id: str,
    step_id: str,
) -> str:
    """
    Generate DataProcessInstance URN for a step execution.

    Args:
        config: Matillion source configuration
        project_id: Matillion project ID
        pipeline_name: Pipeline name
        execution_id: Pipeline execution ID
        step_id: Step ID within the execution

    Returns:
        DataProcessInstance URN
    """
    dpi_id = datahub_guid(
        {
            "platform": MATILLION_PLATFORM,
            "instance": config.platform_instance,
            "env": config.env,
            "project_id": project_id,
            "pipeline_name": pipeline_name,
            "execution_id": execution_id,
            "step_id": step_id,
        }
    )
    return make_data_process_instance_urn(dpi_id)


def build_data_job_custom_properties(
    pipeline: MatillionPipeline, project: MatillionProject
) -> Dict[str, str]:
    custom_properties = {
        "project_id": project.id,
    }

    if hasattr(pipeline, "id") and pipeline.id:
        custom_properties["pipeline_id"] = pipeline.id
    if hasattr(pipeline, "published_time") and pipeline.published_time:
        custom_properties["published_time"] = pipeline.published_time.isoformat()

    return custom_properties


def make_dataset_urn_from_matillion_dataset(dataset: MatillionDatasetInfo) -> str:
    return make_dataset_urn_with_platform_instance(
        platform=dataset.platform,
        name=dataset.name,
        env=dataset.env,
        platform_instance=dataset.platform_instance,
    )


class MatillionUrnBuilder:
    def __init__(self, config: MatillionSourceConfig):
        self.config = config

    def make_project_container_urn(self, project: MatillionProject) -> str:
        return make_container_urn(
            guid=project.id,
        )

    def make_environment_container_urn(
        self, environment: MatillionEnvironment, project: MatillionProject
    ) -> str:
        return make_container_urn(
            guid=f"{project.id}.{environment.name}",
        )

    def make_pipeline_urn(
        self, pipeline: MatillionPipeline, project: MatillionProject
    ) -> str:
        # Use GUID to ensure URN safety with special characters in pipeline names
        flow_id = datahub_guid(
            {
                "platform": MATILLION_PLATFORM,
                "instance": self.config.platform_instance,
                "env": self.config.env,
                "project_id": project.id,
                "pipeline_name": pipeline.name,
            }
        )
        return make_data_flow_urn(
            orchestrator=MATILLION_PLATFORM,
            flow_id=flow_id,
            cluster=self.config.env,
            platform_instance=self.config.platform_instance or project.name,
        )

    def make_data_job_urn(
        self, pipeline: MatillionPipeline, project: MatillionProject
    ) -> str:
        flow_urn = self.make_pipeline_urn(pipeline, project)

        # Use GUID to ensure URN safety with special characters in pipeline names
        job_id = datahub_guid(
            {
                "platform": MATILLION_PLATFORM,
                "instance": self.config.platform_instance,
                "env": self.config.env,
                "project_id": project.id,
                "pipeline_name": pipeline.name,
                "entity_type": "job",  # Distinguish from flow_id
            }
        )

        return make_data_job_urn_with_flow(flow_urn, job_id)

    def make_platform_instance_urn(self) -> Optional[str]:
        if self.config.platform_instance:
            return make_dataplatform_instance_urn(
                platform=MATILLION_PLATFORM,
                instance=self.config.platform_instance,
            )
        return None
